from pytest import mark

from coinflip import randtests_refimpl as randtests

from .examples import *


@mark.parametrize(example_fields, examples)
def test_examples(randtest, bits, statistic_expect, p_expect, kwargs):
    randtest_method = getattr(randtests, randtest)

    statistic, p = randtest_method(bits, **kwargs)

    assert_statistic(statistic, statistic_expect)
    assert_p(p, p_expect)


@mark.parametrize(multi_example_fields, multi_examples)
def test_multi_examples(randtest, bits, expected_statistics, expected_pvalues, kwargs):
    randtest_method = getattr(randtests, randtest)

    statistics, pvalues = randtest_method(bits, **kwargs)

    assert_statistics(statistics, expected_statistics)
    assert_pvalues(pvalues, expected_pvalues)



"""Statistical tests for randomness

This sub-package comprises of implementations of statistical tests laid out in a
comprehensive paper from NIST [1]_ in regards to assessing (pseudo-)random number
generators.

Notes
-----
A copy of the NIST paper can be found at the root of the ``coinflip`` repository as
`SP800-22.pdf <https://github.com/Honno/coinflip/blob/master/SP800-22.pdf>`_.

The test themselves are defined in section 2., "Random Number Generation Tests",
p. 23-62. Further detail of the tests are provided in section 3. "Technical
Descriptions of Tests", p. 63-87.

These tests were implemented in a complimentary program ``sts``, which can be
downloaded from the `NIST website
<https://csrc.nist.gov/projects/random-bit-generation/documentation-and-software>`_.

Note that the paper assumes a great amount of familiarity with certain
concepts in statistics. It also uses some constants and algorithms without any
explanations. Part of the purpose for ``coinflip`` is to "describe" the NIST tests
more wholly than in the paper itself, whilst also reducing the noise of some
non-idiomatic programming conventions used in ``sts``.

.. [1] National Institute of Standards and Technology <Andrew Rukhin, Juan Soto,
   James Nechvatal, Miles Smid, Elaine Barker, Stefan Leigh, Mark Levenson, Mark
   Vangel, David Banks, Alan Heckert, James Dray, San Vo,
   Lawrence E Bassham II>, "A Statistical Test Suite for Random and Pseudorandom
   Number Generators for Cryptographic Applications", *Special Publication
   800-22 Revision 1a*, April 2010.
"""
from typing import Optional
from typing import Tuple

from coinflip import _randtests
from coinflip._randtests.common import exceptions

__all__ = [
    "monobit",
    "frequency_within_block",
    "runs",
    "longest_runs",
    "binary_matrix_rank",
    "spectral",
    "non_overlapping_template_matching",
    "overlapping_template_matching",
    "maurers_universal",
    "linear_complexity",
    "serial",
    "approximate_entropy",
    "cusum",
    "random_excursions",
    "random_excursions_variant",
    "exceptions",
]


def monobit(sequence):
    """Proportion of values is compared to expected 1:1 ratio

    The difference between the occurrences of the two values in the sequence is
    found, and referenced to a hypothetically truly random sequence.

    Parameters
    ----------
    sequence : array-like with two distinct values
        Sequence containing 2 distinct elements

    Returns
    -------
    result : ``MonobitTestResult``
        Dataclass that contains the test's statistic and p-value as well as
        other relevant information gathered.
    """
    return _randtests.monobit(sequence)


def frequency_within_block(sequence, blocksize: Optional[int] = None):
    """Proportion of values per block is compared to expected 1:1 ratio

    The sequence is split into blocks, and the difference between the occurrences
    of the two values in the sequence is found. This is referenced to a
    hypothetically truly random sequence.

    Parameters
    ----------
    sequence : array-like with two distinct values
        Sequence containing 2 distinct elements
    blocksize : ``int``
        Size of the blocks that partition the given sequence

    Returns
    -------
    result : ``FrequencyWithinBlockTestResult``
        Dataclass that contains the test's statistic and p-value as well as
        other relevant information gathered.
    """
    return _randtests.frequency_within_block(sequence, blocksize=blocksize)


def runs(sequence):
    """Number of runs is compared to expected result

    The number of runs (uninterrupted sequence of the same value) is found, and
    referenced to a hypothetically truly random sequence.

    Parameters
    ----------
    sequence : array-like with two distinct values
        Sequence containing 2 distinct elements

    Returns
    -------
    result : ``RunsTestResult``
        Dataclass that contains the test's statistic and p-value as well as
        other relevant information gathered.
    """
    return _randtests.runs(sequence)


def longest_runs(sequence):
    """Longest runs per block is compared to expected result

    The sequence is split into blocks, where the longest number of runs
    (uninterrupted sequence of the same value) in each block is found. This is
    referenced to a hypothetically truly random sequence.

    Parameters
    ----------
    sequence : array-like with two distinct values
        Sequence containing 2 distinct elements

    Returns
    -------
    result : ``LongestRunsTestResult``
        Dataclass that contains the test's statistic and p-value as well as
        other relevant information gathered.
    """
    return _randtests.longest_runs(sequence)


def binary_matrix_rank(sequence, matrix_dimen: Optional[Tuple[int, int]] = None):
    """Independence of neighbouring sequences is compared to expected result

    The sequence is split into matrices, where the rank of each matrix is found
    to determine independence. The counts of different rank bins is referenced
    to a hypothetically truly random sequence.

    Parameters
    ----------
    sequence : array-like with two distinct values
        Sequence containing 2 distinct elements
    matrix_dimen : ``Tuple[int, int]``
        Number of rows and columns in each matrix

    Returns
    -------
    result : ``BinaryMatrixRankTestResult``
        Dataclass that contains the test's statistic and p-value as well as
        other relevant information gathered.
    """
    return _randtests.binary_matrix_rank(sequence, matrix_dimen=matrix_dimen)


def spectral(sequence):
    """Potency of periodic features in sequence is compared to expected result

    The sequence is treated as a signal, which is applied a Fourier transform so
    as to find constituent periodic features. The strength of these features is
    referenced to the expected periodic features present in a hypothetically
    truly random sequence.

    Parameters
    ----------
    sequence : array-like with two distinct values
        Sequence containing 2 distinct elements

    Returns
    -------
    result : ``SpectralTestResult``
        Dataclass that contains the test's statistic and p-value as well as
        other relevant information gathered.

    Raises
    ------
    NonBinaryTruncatedSequenceError
        When odd-lengthed sequence is truncated there is only one distinct value
        present

    """
    return _randtests.spectral(sequence)


def non_overlapping_template_matching(
    sequence, template_size: Optional[int] = None, blocksize: Optional[int] = None
):
    """Matches to template per block is compared to expected result

    The sequence is split into blocks, where the number of non-overlapping
    pattern matches to the template in each block is found. This is referenced
    to the expected mean and variance in matches of a hypothetically truly
    random sequence.

    Parameters
    ----------
    sequence : array-like with two distinct values
        Sequence containing 2 distinct elements
    template_size : ``int``
        Size of all the templates to be generated
    blocksize : ``int``
        Size of the blocks that partition the given sequence

    Returns
    -------
    results : ``NonOverlappingTemplateMatchingMultiTestResult``
        Dataclass that contains the statistics and p-values of all sub-tests, as
        well as other relevant information gathered.
    """
    return _randtests.non_overlapping_template_matching(
        sequence,
        template_size=template_size,
        blocksize=blocksize,
    )


def overlapping_template_matching(
    sequence, template_size: Optional[int] = None, blocksize: Optional[int] = None
):
    """Overlapping matches to template per block is compared to expected result

    The sequence is split into blocks, where the number of overlapping patterns
    matches to the template in each block is found. This is referenced to the
    expected mean and variance in matches of a hypothetically truly random sequence.

    Parameters
    ----------
    sequence : array-like with two distinct values
        Sequence containing 2 distinct elements
    template_size : ``int``
        Size of the template to be generated
    blocksize : ``int``
        Size of the blocks that partition the given sequence

    Returns
    -------
    result : ``OverlappingTemplateMatchingTestResult``
        Dataclass that contains the test's statistic and p-value as well as
        other relevant information gathered.
    """
    return _randtests.overlapping_template_matching(
        sequence,
        template_size=template_size,
        blocksize=blocksize,
    )


def maurers_universal(
    sequence, blocksize: Optional[int] = None, init_nblocks: Optional[int] = None
):
    """Distance between patterns is compared to expected result

    The distinct patterns in an initial sequence are identified, and the
    distances between subsequent occurences of these patterns in a remaining
    sequence are accumulated. The normalised value for the accumulated distances
    is referenced to a hypothetically truly random sequence.

    Parameters
    ----------
    sequence : array-like with two distinct values
        Sequence containing 2 distinct elements
    blocksize : ``int``
        Size of the patterns
    init_nblocks : ``int``
        Number of initial blocks to identify patterns

    Returns
    -------
    result : ``UniversalTestResult``
        Dataclass that contains the test's statistic and p-value as well as
        other relevant information gathered.
    """
    return _randtests.maurers_universal(
        sequence,
        blocksize=blocksize,
        init_nblocks=init_nblocks,
    )


def linear_complexity(sequence, blocksize: Optional[int] = None):
    """LSFRs of blocks is compared to expected length

    The seqience is split into blocks, where the shortest linear feedback shift
    register is found for each block. The difference of each LSFR's length to the
    expected mean length is binned, and is referenced to a hypothetically truly
    random sequence.

    Parameters
    ----------
    sequence : array-like with two distinct values
        Sequence containing 2 distinct elements
    blocksize : ``int``
        Size of the blocks

    Returns
    -------
    results : ``LinearComplexityTestResult``
        Dataclass that contains the test's statistic and p-value as well as
        other relevant information gathered.
    """
    return _randtests.linear_complexity(sequence, blocksize=blocksize)


def serial(sequence, blocksize: Optional[int] = None):
    """Proportion of all overlapping patterns is compared to expected uniformity

    The number of overlapping pattern matches for each distinct pattern is
    found. This is referenced to a hypothetically truly random sequence.

    Parameters
    ----------
    sequence : array-like with two distinct values
        Sequence containing 2 distinct elements
    blocksize : ``int``
        Size of the patterns

    Returns
    -------
    results : ``SerialMultiTestResult``
        Dataclass that contains the statistics and p-values of all sub-tests, as
        well as other relevant information gathered.
    """
    return _randtests.serial(sequence, blocksize=blocksize)


def approximate_entropy(sequence, blocksize: Optional[int] = None):
    """Approximate entropy of sequence is compared to expected result

    The approximate entropy of the sequence is found and referenced to a truly
    random sequence.

    Parameters
    ----------
    sequence : array-like with two distinct values
        Sequence containing 2 distinct elements
    blocksize : ``int``
        Size of the patterns

    Returns
    -------
    results : ``ApproximateEntropyTestResult``
        Dataclass that contains the test's statistic and p-value as well as
        other relevant information gathered.
    """
    return _randtests.approximate_entropy(sequence, blocksize=blocksize)


def cusum(sequence, reverse: bool = False):
    """Furthest detour in a randomn walk is compared to expected result

    The sequence is treated as a random walk, where the furthest detour from the
    axis is identified and referenced to a hypothetically truly random sequence.

    Parameters
    ----------
    sequence : array-like with two distinct values
        Sequence containing 2 distinct elements
    reverse : ``bool``
        Cumulate sums from the end of the sequence first.

    Returns
    -------
    results : ``CusumTestResult``
        Dataclass that contains the test's statistic and p-value as well as
        other relevant information gathered.
    """
    return _randtests.cusum(sequence, reverse=reverse)


def random_excursions(sequence):
    """Frequency of states per cycle in a random walk is compared to expected results

    The sequence is treated as a random walk, where the frequency of states -4
    to 4 in each cycle is found. This is referenced to a hypothetically truly
    random sequence.

    Parameters
    ----------
    sequence : array-like with two distinct values
        Sequence containing 2 distinct elements

    Returns
    -------
    results : ``RandomExcursionsMultiTestResult``
        Dataclass that contains the statistics and p-values of all sub-tests, as
        well as other relevant information gathered.
    """
    return _randtests.random_excursions(sequence)


def random_excursions_variant(sequence):
    """Proportion of states in a random walk is compared to expected results

    The sequence is treated as a random walk, where the occurrences of states -9
    to 9 is founded and referenced to a hypothetically truly random sequence.

    Parameters
    ----------
    sequence : array-like with two distinct values
        Sequence containing 2 distinct elements

    Returns
    -------
    results : ``RandomExcursionsVariantMultiTestResult``
        Dataclass that contains the statistics and p-values of all sub-tests, as
        well as other relevant information gathered.
    """
    return _randtests.random_excursions_variant(sequence)



from collections import defaultdict
from dataclasses import dataclass
from math import floor
from math import log
from math import log2

import pandas as pd
from scipy.special import gammaincc

from coinflip._randtests.common.core import *
from coinflip._randtests.common.result import TestResult
from coinflip._randtests.common.testutils import slider
from coinflip._randtests.common.typing import Integer

__all__ = ["approximate_entropy"]


@randtest()
def approximate_entropy(series, heads, tails, ctx, blocksize=None):
    n = len(series)

    if not blocksize:
        blocksize = max(floor(log2(n)) - 5 - 1, 2)

    set_task_total(ctx, (n + 2) * 2 + 1)

    failures = check_recommendations(
        ctx, {"blocksize < ⌊log2(n)⌋ - 5": blocksize < floor(log2(n)) - 5}
    )

    phis = []
    for template_size in [blocksize, blocksize + 1]:
        head = series[: template_size - 1]
        ouroboros = pd.concat([series, head])

        permutation_counts = defaultdict(int)
        for window_tup in slider(ouroboros, template_size):
            permutation_counts[window_tup] += 1

            advance_task(ctx)

        normalised_counts = []
        for count in permutation_counts.values():
            normcount = count / n
            normalised_counts.append(normcount)

        advance_task(ctx)

        phi = sum(normcount * log(normcount) for normcount in normalised_counts)
        phis.append(phi)

        advance_task(ctx)

    approx_entropy = phis[0] - phis[1]
    chi2 = 2 * n * (log(2) - approx_entropy)
    p = gammaincc(2 ** (blocksize - 1), chi2 / 2)

    advance_task(ctx)

    return ApproximateEntropyTestResult(heads, tails, failures, chi2, p, blocksize)


@dataclass
class ApproximateEntropyTestResult(TestResult):
    blocksize: Integer

    def _render(self):
        yield self._pretty_result("chi-square")



from collections import Counter
from dataclasses import dataclass
from functools import lru_cache
from math import erfc
from math import sqrt
from operator import attrgetter
from typing import List
from typing import NamedTuple

import altair as alt
import numpy as np
import pandas as pd
from rich.text import Text
from scipy.special import gammaincc
from scipy.stats import halfnorm

from coinflip._randtests.common.core import *
from coinflip._randtests.common.result import TestResult
from coinflip._randtests.common.result import make_testvars_table
from coinflip._randtests.common.result import smartround
from coinflip._randtests.common.testutils import blocks
from coinflip._randtests.common.typing import Face
from coinflip._randtests.common.typing import Integer

__all__ = ["monobit", "frequency_within_block"]


# ------------------------------------------------------------------------------
# Frequency (Monobit) Test


class FaceCount(NamedTuple):
    value: Face
    count: Integer


class FaceCounts(NamedTuple):
    heads: FaceCount
    tails: FaceCount

    @property
    @lru_cache()
    def max(self):
        return max(*self, key=attrgetter("count"))

    @property
    @lru_cache()
    def min(self):
        return min(*self, key=attrgetter("count"))

    @classmethod
    def from_series(cls, series, heads, tails):
        heads = FaceCount(heads, series[heads])
        tails = FaceCount(tails, series[tails])
        return cls(heads, tails)


@randtest()
def monobit(series, heads, tails, ctx):
    n = len(series)

    set_task_total(ctx, 2)

    failures = check_recommendations(ctx, {"n ≥ 100": n >= 100})

    counts = FaceCounts.from_series(series.value_counts(), heads, tails)

    advance_task(ctx)

    diff = counts.max.count - counts.min.count
    normdiff = diff / sqrt(n)
    p = erfc(normdiff / sqrt(2))

    advance_task(ctx)

    return MonobitTestResult(heads, tails, failures, normdiff, p, n, counts, diff)


@dataclass
class MonobitTestResult(TestResult):
    n: Integer
    counts: FaceCounts
    diff: Integer

    def _render(self):
        yield self._pretty_result("normalised diff")

        counts = make_testvars_table("value", "count")
        counts.add_row(str(self.counts.max.value), str(self.counts.max.count))
        counts.add_row(str(self.counts.min.value), str(self.counts.min.count))
        yield counts

    def plot_counts(self):
        df = pd.DataFrame(
            {
                "Value": [self.counts.max.value, self.counts.min.value],
                "Count": [self.counts.max.count, self.counts.min.count],
            }
        )
        df["Value"] = df["Value"].astype("category")

        chart = (
            alt.Chart(df)
            .mark_bar()
            .encode(
                alt.X("Value"),
                alt.Y("Count", axis=alt.Axis(tickMinStep=1)),
                tooltip="Count",
            )
            .properties(
                title=f"Counts of {self.counts.max.value} and {self.counts.min.value}"
            )
        )

        return chart

    def plot_refdist(self):
        mean = 0
        variance = 1
        deviation = sqrt(variance)

        xlim = max(4 * deviation, self.statistic + deviation)
        x = np.linspace(0, xlim)
        y = halfnorm.pdf(x, mean, deviation)
        x_stat = np.linspace(self.statistic, xlim)
        y_stat = halfnorm.pdf(x_stat, mean, deviation)

        dist = pd.DataFrame({"x": x, "y": y})
        dist_stat = pd.DataFrame({"x": x_stat, "y": y_stat})

        chart_dist = (
            alt.Chart(dist)
            .mark_area(opacity=0.3)
            .encode(
                alt.X(
                    "x", axis=alt.Axis(title="Difference between counts (normalised)")
                ),
                alt.Y(
                    "y",
                    axis=alt.Axis(title="Probability density"),
                    scale=alt.Scale(domain=(0, 1)),
                ),
            )
            .properties(title="Proability density of count differences")
        )
        chart_stat = (
            alt.Chart(dist_stat)
            .mark_area()
            .encode(
                x="x",
                y="y",
            )
        )
        chart = chart_dist + chart_stat

        return chart


# ------------------------------------------------------------------------------
# Frequency within Block Test


@randtest(min_n=8)
def frequency_within_block(series, heads, tails, ctx, blocksize=None):
    n = len(series)

    # TODO - smarter defaults
    #      - meet 0.01 * n recommendation
    if not blocksize:
        blocksize = 8

    nblocks = n // blocksize

    set_task_total(ctx, nblocks + 3)

    failures = check_recommendations(
        ctx,
        {
            "n ≥ 100": n >= 100,
            "blocksize ≥ 20": blocksize >= 20,
            "blocksize > 0.01 * n": blocksize > 0.01 * n,
            "nblocks < 100": nblocks < 100,
        },
    )

    advance_task(ctx)

    counts = []
    for block in blocks(series, blocksize):
        matches = block == heads
        count = matches.sum()
        counts.append(count)

        advance_task(ctx)

    proportions = (count / blocksize for count in counts)
    deviations = [prop - 1 / 2 for prop in proportions]

    advance_task(ctx)

    # TODO figure out the chi-square test being used
    statistic = 4 * blocksize * sum(x ** 2 for x in deviations)
    p = gammaincc(nblocks / 2, statistic / 2)

    advance_task(ctx)

    return FrequencyWithinBlockTestResult(
        heads,
        tails,
        failures,
        statistic,
        p,
        blocksize,
        nblocks,
        counts,
    )


@dataclass
class FrequencyWithinBlockTestResult(TestResult):
    blocksize: Integer
    nblocks: Integer
    counts: List[Integer]

    def _render(self):
        yield self._pretty_result("chi-square")

        yield TestResult._pretty_inputs(
            ("blocksize", self.blocksize), ("nblocks", self.nblocks)
        )

        title = Text.assemble("count of ", (str(self.heads), "bold"), " per block")

        count_expect = self.blocksize / 2
        f_count_expect = smartround(count_expect)
        caption = f"expected count {f_count_expect}"

        count_nblocks = Counter(self.counts)
        table = make_testvars_table("count", "nblocks", title=title, caption=caption)
        for count in range(self.blocksize + 1):
            nblocks = count_nblocks[count]
            table.add_row(str(count), str(nblocks))

        yield table

    # TODO delete this when jinja2 template and altair plotting methods are done
    # def _report(self):
    #     occurfig, occurax = plt.subplots()

    #     x_axis = [i * self.blocksize for i in range(len(self.counts))]

    #     occurax.set_ylim([0, self.blocksize])
    #     occurax.bar(x_axis, self.counts, width=self.blocksize * 0.9, align="edge")
    #     occurax.axhline(self.blocksize / 2, color="black")

    #     return [
    #         f"p={self.p3f()}",
    #         occurfig,
    #         plot_chi2(self.statistic, df=self.nblocks),
    #         plot_gammaincc(self.statistic / 2, self.nblocks / 2),
    #     ]



from datetime import timedelta

from hypothesis import settings
from pytest import mark

settings.register_profile(
    "fast", max_examples=2, stateful_step_count=4, deadline=timedelta(minutes=2)
)

settings.register_profile(
    "debug", max_examples=1, stateful_step_count=1, deadline=timedelta(minutes=5),
)


def pytest_addoption(parser):
    parser.addoption(
        "--run-slow", action="store_true", default=False, help="Run slow tests"
    )


def pytest_configure(config):
    config.addinivalue_line("markers", "slow: mark test as slow to run")

    if not config.getoption("--run-slow"):
        settings.load_profile("fast")


def pytest_collection_modifyitems(config, items):
    if not config.getoption("--run-slow"):
        skip_slow = mark.skip(reason="Needs --run-slow option to run")
        for item in items:
            if "slow" in item.keywords:
                item.add_marker(skip_slow)



import webbrowser
from datetime import datetime
from pathlib import Path

import pandas as pd
from click import Choice
from click import Path as Path_
from click import argument
from click import group
from click import option
from rich.text import Text

from coinflip import generators
from coinflip._randtests.common.exceptions import NonBinarySequenceError
from coinflip._randtests.common.exceptions import TestError
from coinflip.cli import console
from coinflip.cli.parsing import DataParsingError
from coinflip.cli.parsing import *
from coinflip.cli.pprint import *
from coinflip.cli.report import *
from coinflip.cli.runner import *
from coinflip.randtests import __all__ as randtest_names

__all__ = ["run", "example_run", "read", "report"]


# TODO extend Choice to use print_error and newline-delimit lists
test_choice = Choice(randtest_names)


CONTEXT_SETTINGS = {"help_option_names": ["-h", "--help"]}


@group(context_settings=CONTEXT_SETTINGS)
def main():
    """Randomness tests for RNG output.

    Randomness tests can be ran on the your binary data via the run command.
    Informational web documents explaining the test results can be produced via
    the report command.
    """


@main.command()
@argument("data", type=Path_(exists=True))
@argument("out", type=Path_(), required=False, metavar="OUT")
@option(
    "-b",
    "--binary",
    is_flag=True,
    flag_value=True,
    help="Read DATA as a raw binary file.",
)
def run(data, out, binary):
    """Run randomness tests on DATA and write results to OUT.

    DATA is a newline-delimited text file which contains output of a random
    number generator.

    Individual results of each test are printed as they come. Once they are
    all finished, the results are written to OUT.

    The results saved in OUT can be printed again via the read command. OUT can
    also be used to generate an informational web document via the report
    command.
    """
    if not binary:
        try:
            series = parse_text(data)
        except (DataParsingError, NonBinarySequenceError) as e:
            print_error(e)
            exit(1)

    else:
        series = parse_binary(data)

    print_series(series)

    results = {}
    for name, result, e in run_all_tests(series):
        if not e:
            results[name] = result

    if out:
        path = Path(out)
        # TODO check if not pickle at end
    else:
        timestamp = datetime.now()
        f_timestamp = timestamp.strftime("%b%d_%H%M%S")
        path = Path(f"results_{f_timestamp}.pickle")

    store_results(series, results, path)

    if not out:
        console.print("")
        console.print(f"Results saved to {path}")


@main.command()
@option(
    "-e",
    "--example",
    type=Choice(generators.__all__),
    default="python",
    help="Example binary output to use.",
    metavar="<example>",
)
@option("-n", "--length", type=int, default=512, help="Length of binary output.")
@option(
    "-t",
    "--test",
    type=test_choice,
    help="Specify single test to run on data.",
    metavar="<test>",
)
def example_run(example, length, test):
    """Run randomness tests on automatically generated data."""
    generator_func = getattr(generators, example)
    generator = generator_func()

    series = pd.Series(next(generator) for _ in range(length))
    series = series.infer_objects()
    print_series(series)
    console.print()

    if not test:
        for name, result, e in run_all_tests(series):
            pass

    else:
        try:
            run_test(series, test)
        except TestError:
            exit(1)


@main.command()
@argument("results", type=Path_(exists=True))
def read(results):
    """Print test results."""
    report = load_results(results)
    print_series(report.series)
    console.print("")
    print_results(report.results)


@main.command()
@argument("results", type=Path_(exists=True))
@argument("out", type=Path_(), required=False, metavar="OUT")
def report(results, out):
    """Generate an informational web document from results."""
    report = load_results(results)

    if out:
        path = Path(out)
        # TODO check if not html at end
    else:
        timestamp = datetime.now()
        f_timestamp = timestamp.strftime("%b%d_%H%M%S")
        path = Path(f"report_{f_timestamp}.html")

    write_report_doc(report, path)

    if not out:
        console.print(f"Report saved to {path}")

    abs_path = path.absolute()
    url = abs_path.as_uri()
    url_msg = Text.assemble("View report at ", (str(abs_path), f"link {url}"))

    console.print(url_msg)
    webbrowser.open(url)



from dataclasses import dataclass
from io import StringIO
from typing import Any
from typing import Dict
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Tuple
from typing import Union

from rich import box
from rich.console import Console
from rich.console import ConsoleRenderable
from rich.console import RenderableType
from rich.console import RenderGroup
from rich.console import render_group
from rich.measure import Measurement
from rich.measure import measure_renderables
from rich.padding import Padding
from rich.segment import Segment
from rich.table import Table
from rich.text import Text
from typing_extensions import get_args

from coinflip._randtests.common.core import make_failures_msg
from coinflip._randtests.common.pprint import make_warning
from coinflip._randtests.common.typing import Face
from coinflip._randtests.common.typing import Float
from coinflip._randtests.common.typing import Integer

__all__ = [
    "BaseTestResult",
    "TestResult",
    "MultiTestResult",
    "make_testvars_table",
    "make_chisquare_table",
    "smartround",
]


@dataclass
class BaseTestResult(ConsoleRenderable):
    heads: Face
    tails: Face
    failures: List[str]

    def _render(self) -> Iterator[RenderableType]:
        pass

    def __rich_console__(self, console, options):
        newline = Segment.line()
        *renderables, last_renderable = self.renderables

        for renderable in renderables:
            yield renderable
            yield newline
        yield last_renderable

    def __rich_measure__(self, console, max_width):
        return measure_renderables(console, self.renderables, max_width)

    @property
    def renderables(self) -> List[RenderableType]:
        renderables = list(self._render())

        if self.failures:
            msg = make_failures_msg(self.failures)
            f_failures = make_warning(msg)
            renderables.insert(0, f_failures)

        return renderables

    def print(self, **kwargs):
        """Prints results contents to notebook or terminal environment"""
        console = Console()
        console.print(self, **kwargs)

    def __str__(self):
        # Mocks file as the stdout of a Rich Console
        buf = StringIO()
        console = Console(file=buf, force_jupyter=False)
        console.print(self)

        return buf.getvalue()

    def _pretty_inputs(
        self, *name_value_pairs: Iterable[Union[Integer, Float]]
    ) -> RenderGroup:
        title = "test input"
        if len(name_value_pairs) > 1:
            title += "s"
        return make_testvars_list(title, *name_value_pairs)


class PrettyResultMixin:
    def _pretty_result(self, stat_varname="statistic", prefix=None) -> RenderGroup:
        return make_testvars_list(
            f"{prefix} result" if prefix else "result",
            (stat_varname, smartround(self.statistic)),
            ("p-value", round(self.p, 3)),
        )


@dataclass(unsafe_hash=True)
class TestResult(BaseTestResult, PrettyResultMixin):
    statistic: Union[Integer, Float]
    p: Float


@dataclass(unsafe_hash=True)
class SubTestResult(PrettyResultMixin):
    statistic: Union[Integer, Float]
    p: Float


@dataclass(unsafe_hash=True)
class MultiTestResult(BaseTestResult):
    results: Dict[Any, SubTestResult]

    # TODO lru_cache() the properties once hashing works nicely

    @property
    def statistics(self) -> List[Union[Integer, float]]:
        return [result.statistic for result in self.results.values()]

    @property
    def pvalues(self) -> List[float]:
        return [result.p for result in self.results.values()]

    @property
    def min(self):
        items = iter(self.results.items())
        min_feature, min_result = next(items)
        for feature, result in items:
            if result.p < min_result.p:
                min_feature = feature
                min_result = result

        return min_feature, min_result

    def _pretty_feature(self, result: SubTestResult) -> Text:
        pass

    def _render_sub(self, result: SubTestResult) -> Iterator[RenderableType]:
        pass

    def _results_table(self, feature_varname: str, stat_varname: str) -> Table:
        min_feature, min_result = self.min

        table = make_testvars_table(
            feature_varname, stat_varname, "p", title="sub-test results"
        )
        for feature, result in self.results.items():
            f_feature = self._pretty_feature(result)
            if feature == min_feature:
                f_feature += Text("*", style="blue")

            f_statistic = str(round(result.statistic, 3))
            f_p = str(round(result.p, 3))

            table.add_row(f_feature, f_statistic, f_p)

        min_title = Text(f"sub-test result for {feature_varname} ", style="italic")
        min_title.append(self._pretty_feature(min_result))

        f_result = Table.grid(padding=(1))
        for renderable in self._render_sub(min_result):
            f_result.add_row(renderable)

        titled_result = RenderGroup(min_title, "", f_result)

        example = Table.grid()
        example.add_row(Text.assemble(("*", "bold blue"), " "), titled_result)

        meta_table = Table.grid()
        padded_example = Padding(example, (4, 0, 0, 3))
        meta_table.add_row(table, padded_example)

        return meta_table


def make_chisquare_table(
    title: Union[str, Text],
    feature: Union[Text, str],
    classes: Iterable[Any],
    expected_occurences: Iterable[Union[Integer, float]],
    actual_occurences: Iterable[Union[Integer, float]],
    **kwargs,
) -> Table:
    f_table = make_testvars_table(
        feature, "expect", "actual", "diff", title=title, **kwargs
    )

    table = zip(classes, expected_occurences, actual_occurences)
    for class_, nblocks_expect, nblocks in table:
        f_class = str(class_)
        f_nblocks_expect = str(smartround(nblocks_expect))
        f_nblocks = str(nblocks)

        diff = nblocks - nblocks_expect
        f_diff = str(smartround(diff, 1))

        f_table.add_row(f_class, f_nblocks_expect, f_nblocks, f_diff)

    return f_table


def make_testvars_table(*columns, box=box.SQUARE, **kwargs) -> Table:
    table = OverflowTable(box=box, **kwargs)
    table.add_column(columns[0], justify="left")
    for col in columns[1:]:
        table.add_column(col, justify="right")

    return table


@render_group(fit=True)
def make_testvars_list(
    title: str, *varname_value_pairs: Tuple[str, Union[Integer, float]]
) -> RenderGroup:
    yield Text(title, style="bold")

    varnames, values = zip(*varname_value_pairs)

    varname_maxlen = max(len(varname) for varname in varnames)
    f_varnames = [pad_right(varname, varname_maxlen) for varname in varnames]

    f_values = align_nums(values)

    for f_varname, f_value in zip(f_varnames, f_values):
        yield f"  {f_varname}  {f_value}"


def smartround(num: Union[Integer, Float], ndigits=1) -> Union[int, float]:
    """Round number only if it's a float"""
    if isinstance(num, get_args(Integer)):
        return int(num)
    else:
        num = float(num)
        if num.is_integer():
            return int(num)
        else:
            return round(num, ndigits)


# ------------------------------------------------------------------------------
# Helpers


class OverflowTable(Table):
    def __init__(self, *args, title=None, caption=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.overflow_title = title
        self.overflow_caption = caption

    def __rich_console__(self, console, options):
        # TODO think about the ramifications of what I'm doing here (tests?)
        table = super().__rich_console__(console, options)

        max_width = options.max_width
        if self.width is not None:
            max_width = self.width
        if self.box:
            max_width -= len(self.columns) - 1
            if self.show_edge:
                max_width -= 2
        widths = self._calculate_column_widths(console, max_width)
        table_width = sum(widths) + self._extra_width

        render_options = options.update(width=table_width)

        def overflow_render_annotation(text: Union[str, Text], style, justify="center"):
            if isinstance(text, str):
                render_text = console.render_str(text, style=style)
            else:
                text.stylize(style)
                render_text = text

            if len(render_text) > table_width:
                return console.render(render_text, options=options)
            else:
                return console.render(
                    render_text, options=render_options.update(justify=justify)
                )

        if self.overflow_title:
            yield from overflow_render_annotation(
                self.overflow_title, "table.title", justify=self.title_justify
            )

        yield from table

        if self.overflow_caption:
            yield from overflow_render_annotation(
                self.overflow_caption, "table.caption", justify=self.caption_justify
            )

    def __rich_measure__(self, console, max_width):
        _min_width, _max_width = super().__rich_measure__(console, max_width)

        widths = [_min_width]
        if self.overflow_title:
            widths.append(len(self.overflow_title))
        if self.overflow_caption:
            widths.append(len(self.overflow_caption))

        min_width = max(widths)
        max_width = max(_max_width, min_width)

        return Measurement(min_width, max_width)


def align_nums(nums):
    f_nums = [str(smartround(num, 3)) for num in nums]

    dot_positions = []
    for f_num in f_nums:
        pos = f_num.find(".")
        if pos == -1:
            pos = len(f_num)
        dot_positions.append(pos)
    maxpos = max(dot_positions)

    f_aligned_nums = []
    for f_num, pos in zip(f_nums, dot_positions):
        left_pad = " " * (maxpos - pos)
        f_num = left_pad + f_num
        f_aligned_nums.append(f_num)

    return f_aligned_nums


def pad_right(string, maxlen):
    nspaces = maxlen - len(string)
    f_string = string + " " * nspaces

    return f_string



from collections import Counter
from collections import defaultdict
from dataclasses import dataclass
from itertools import product
from math import ceil
from math import exp
from math import floor
from math import isclose
from math import log2
from math import sqrt
from typing import List
from typing import Tuple

from rich.text import Text
from scipy.special import gammaincc
from scipy.special import hyp1f1
from scipy.stats import chisquare

from coinflip._randtests.common.collections import defaultlist
from coinflip._randtests.common.core import *
from coinflip._randtests.common.pprint import pretty_subseq
from coinflip._randtests.common.result import MultiTestResult
from coinflip._randtests.common.result import SubTestResult
from coinflip._randtests.common.result import TestResult
from coinflip._randtests.common.result import make_chisquare_table
from coinflip._randtests.common.result import make_testvars_table
from coinflip._randtests.common.testutils import blocks
from coinflip._randtests.common.testutils import rawblocks
from coinflip._randtests.common.testutils import slider
from coinflip._randtests.common.typing import Face
from coinflip._randtests.common.typing import Float
from coinflip._randtests.common.typing import Integer

__all__ = ["non_overlapping_template_matching", "overlapping_template_matching"]


# ------------------------------------------------------------------------------
# Non-overlapping Template Matching Test


@randtest()
def non_overlapping_template_matching(
    series, heads, tails, ctx, template_size=None, blocksize=None,
):
    n = len(series)

    if not blocksize:
        blocksize = max(ceil(0.01 * n), 6)
        if blocksize % 2 != 0:
            blocksize -= 1
    nblocks = n // blocksize

    if not template_size:
        template_size = max(min(blocksize // 3, 9), 2)

    nblocks_sub = blocksize // template_size
    set_task_total(ctx, 1 + nblocks * (nblocks_sub + 1) + 1)

    failures = check_recommendations(
        ctx,
        {
            "n ≥ 100": n >= 100,
            "template_size = 9 or 10": template_size == 9 or template_size == 10,
            "blocksize > 0.01 * n": blocksize > 0.01 * n,
            "nblocks ≤ 100": nblocks <= 100,  # TODO same thing as above?
            "nblocks = ⌊n / blocksize⌋": nblocks == n // blocksize,
        },
    )

    matches_expect = (blocksize - template_size + 1) / 2 ** template_size
    variance = blocksize * (
        (1 / 2 ** template_size) - ((2 * template_size - 1)) / 2 ** (2 * template_size)
    )

    advance_task(ctx)

    template_block_matches = defaultdict(lambda: defaultlist(int))
    for i, block in enumerate(blocks(series, blocksize)):
        matches = defaultdict(int)

        for window_tup in rawblocks(block, template_size):
            matches[window_tup] += 1

            advance_task(ctx)

        for template, matches in matches.items():
            template_block_matches[template][i] = matches

        advance_task(ctx)

    results = {}
    for template in product([heads, tails], repeat=template_size):
        block_matches = template_block_matches[template][:nblocks]
        match_diffs = [matches - matches_expect for matches in block_matches]

        statistic = sum(diff ** 2 / variance for diff in block_matches)
        p = gammaincc(nblocks / 2, statistic / 2)

        results[template] = NonOverlappingTemplateMatchingSubTestResult(
            statistic, p, template, block_matches, match_diffs,
        )

    advance_task(ctx)

    return NonOverlappingTemplateMatchingMultiTestResult(
        heads,
        tails,
        failures,
        results,
        template_size,
        blocksize,
        nblocks,
        matches_expect,
        variance,
    )


@dataclass(unsafe_hash=True)
class NonOverlappingTemplateMatchingSubTestResult(SubTestResult):
    template: Tuple[Face, ...]
    block_matches: List[Integer]
    match_diffs: List[Float]


@dataclass(unsafe_hash=True)
class NonOverlappingTemplateMatchingMultiTestResult(MultiTestResult):
    template_size: Integer
    blocksize: Integer
    nblocks: Integer
    matches_expect: Float
    variance: Float

    def _pretty_feature(self, result: NonOverlappingTemplateMatchingSubTestResult):
        f_template = pretty_subseq(result.template, self.heads, self.tails)

        return f_template

    # TODO q value
    def _render(self):
        yield self._pretty_inputs(
            ("blocksize", self.blocksize), ("nblocks", self.nblocks),
        )

        yield self._results_table("template", "χ²")

    def _render_sub(self, result: NonOverlappingTemplateMatchingSubTestResult):
        yield result._pretty_result("chi-square")

        title = Text("matches of ")
        title.append(pretty_subseq(result.template, self.heads, self.tails))
        title.append(" per block")

        f_matches_expect = round(self.matches_expect, 1)
        caption = f"expected {f_matches_expect} matches"

        matches_count = Counter(result.block_matches)
        table = sorted(matches_count.items())
        f_table = make_testvars_table(
            "matches", "nblocks", title=title, caption=caption
        )
        for matches, nblocks in table:
            f_table.add_row(str(matches), str(nblocks))

        yield f_table


# ------------------------------------------------------------------------------
# Overlapping Template Matching Test


matches_ceil = 5


# TODO Review paper "Correction of Overlapping Template Matching Test Included in
#                    NIST Randomness Test Suite"
@randtest()  # TODO appropiate min input
def overlapping_template_matching(
    series, heads, tails, ctx, template_size=None, blocksize=None, df=5
):
    n = len(series)

    if not blocksize:
        blocksize = floor(sqrt(n))
    nblocks = n // blocksize

    if not template_size:
        template_size = min(max(floor(sqrt(blocksize)), 2), 12)
    template = [heads for _ in range(template_size)]

    lambda_ = (blocksize - template_size + 1) / 2 ** template_size
    eta = lambda_ / 2

    first_prob = exp(-eta)
    probabilities = [first_prob]
    for matches in range(1, matches_ceil):
        prob = ((eta * exp(-2 * eta)) / 2 ** matches) * hyp1f1(matches + 1, 2, eta)
        probabilities.append(prob)
    last_prob = 1 - sum(probabilities)
    probabilities.append(last_prob)

    set_task_total(ctx, 1 + nblocks + 2)

    failures = check_recommendations(
        ctx,
        {
            "n ≥ 288": n >= 288,
            "n ≥ nblocks * blocksize": n >= nblocks * blocksize,
            "nblocks * min(probabilities) > df": nblocks * min(probabilities) > df,
            "λ ≈ 2": isclose(lambda_, 2),
            "len(template) ≈ log2(nblocks)": isclose(template_size, log2(nblocks)),
            "df ≈ 2 * λ": isclose(template_size, 2 * lambda_),
        },
    )

    expected_tallies = [prob * nblocks for prob in probabilities]

    advance_task(ctx)

    block_matches = []
    for block in blocks(series, blocksize):
        matches = 0

        for window_tup in slider(block, template_size):
            if all(x == y for x, y in zip(window_tup, template)):
                matches += 1

        advance_task(ctx)

        block_matches.append(matches)

    tallies = [0 for _ in range(matches_ceil + 1)]
    for matches in block_matches:
        i = min(matches, matches_ceil)
        tallies[i] += 1

    advance_task(ctx)

    statistic, p = chisquare(tallies, expected_tallies)

    advance_task(ctx)

    return OverlappingTemplateMatchingTestResult(
        heads,
        tails,
        failures,
        statistic,
        p,
        template_size,
        blocksize,
        nblocks,
        template,
        expected_tallies,
        tallies,
    )


@dataclass
class OverlappingTemplateMatchingTestResult(TestResult):
    template_size: Integer
    blocksize: Integer
    nblocks: Integer
    template: Tuple[Face, ...]
    expected_tallies: List[Integer]
    tallies: List[Integer]

    def _render(self):
        yield self._pretty_result("chi-square")

        yield TestResult._pretty_inputs(
            ("template size", self.template_size),
            ("blocksize", self.blocksize),
            ("nblocks", self.nblocks),
        )

        title = Text("matches of ")
        title.append(pretty_subseq(self.template, self.heads, self.tails))
        title.append(" per block")

        f_nmatches = [f"{x}" for x in range(matches_ceil + 1)]
        f_nmatches[-1] = f"{f_nmatches[-1]}+"

        table = make_chisquare_table(
            title, "matches", f_nmatches, self.expected_tallies, self.tallies,
        )

        yield table



"""Generators of binary sequences

Methods infinitely generate ``0`` and ``1`` integers to represent a binary sequence."""
from math import sqrt
from random import getrandbits

__all__ = ["python", "primes"]


def python():
    """Generates random bits using python's ``random`` module

    Yields
    ------
    bit : ``0`` or `1`
        Random bit

    See Also
    --------
    random.getrandbits : Method used to generate bits
    """
    while True:
        yield getrandbits(1)


def numbers(start=1):
    """Generates natural numbers

    Parameters
    ----------
    start : ``int``, default ``1``
        Starting number to count up from

    Yields
    ------
    n : ``int``
        Natural number
    """
    n = start
    while True:
        yield n
        n += 1


def primes():
    """Generates bits representing if natural numbers are prime

    Yields
    ------
    bit: ``0`` or ``1``
        Whether next number is prime: ``0`` represents number is a composite (i.e.
        not a prime), ``1`` represents number is a prime.
    """
    yield 0  # n = 1
    yield 1  # n = 2

    for n in numbers(3):
        if n % 2 == 0:
            yield 0
        else:
            for divisor in range(3, int(sqrt(n)) + 1, 2):
                if n % divisor == 0:
                    yield 0
            else:
                yield 1



"""Methods used to interact with the _randtests subpackage."""
from functools import wraps
from shutil import get_terminal_size
from typing import Callable
from typing import Dict
from typing import Iterator
from typing import Tuple

import pandas as pd
from rich import box
from rich.progress import BarColumn
from rich.progress import Progress
from rich.progress import TextColumn
from rich.progress import TimeRemainingColumn
from rich.rule import Rule
from rich.table import Table
from rich.text import Text

from coinflip import _randtests
from coinflip._randtests.common.exceptions import NonBinarySequenceError
from coinflip._randtests.common.exceptions import TestError
from coinflip._randtests.common.result import BaseTestResult
from coinflip._randtests.common.result import MultiTestResult
from coinflip._randtests.common.result import TestResult
from coinflip.cli import console
from coinflip.cli.pprint import print_error
from coinflip.cli.pprint import print_warning

__all__ = [
    "list_tests",
    "TestNotFoundError",
    "run_test",
    "run_all_tests",
    "print_results",
]


SIGLEVEL = 0.01

f_randtest_names = {
    "monobit": "Frequency (Monobit) Test",
    "frequency_within_block": "Frequency within Block Test",
    "runs": "Runs Test",
    "longest_runs": "Longest Runs in Block Test",
    "binary_matrix_rank": "Matrix Rank Test",
    "spectral": "Discrete Fourier Transform (Spectral) Test",
    "non_overlapping_template_matching": "Non-Overlapping Template Matching Test",
    "overlapping_template_matching": "Overlapping Template Matching Test",
    "maurers_universal": "Maurer's Universal Test",
    "linear_complexity": "Linear Complexity Test",
    "serial": "Serial Test",
    "approximate_entropy": "Approximate Entropy Test",
    "cusum": "Cumulative Sums (Cusum) Test",
    "random_excursions": "Random Excursions Test",
    "random_excursions_variant": "Random Excursions Variant Test",
}


f_randtest_abbreviations = {
    "monobit": "Monobit",
    "frequency_within_block": "Block Freq.",
    "runs": "Runs",
    "longest_runs": "Longest Runs",
    "binary_matrix_rank": "Matrix",
    "spectral": "Spectral",
    "non_overlapping_template_matching": "Non-Overlap.",
    "overlapping_template_matching": "Overlapping",
    "maurers_universal": "Universal",
    "linear_complexity": "Complexity",
    "serial": "Serial",
    "approximate_entropy": "Entropy",
    "cusum": "Cusum",
    "random_excursions": "Excursions",
    "random_excursions_variant": "Excur. Var.",
}


def binary_check(func):
    """Decorator to check if series comprises of binary values"""

    @wraps(func)
    def wrapper(series, *args, **kwargs):
        if series.nunique() != 2:
            raise NonBinarySequenceError()

        return func(series, *args, **kwargs)

    return wrapper


def print_randtest_name(randtest_name: str, color: str):
    """Pretty print the randtest's name"""
    size = get_terminal_size()
    ncols = min(size.columns, 80)

    randtest_fname = f_randtest_names[randtest_name]

    color = f"bright_{color}"

    rule = Rule(randtest_fname, style=color)
    console.print(rule, width=ncols)


def list_tests() -> Iterator[Tuple[str, Callable]]:
    """List all available statistical tests

    Yields
    ------
    randtest_name : ``str``
        Name of statistical test
    randtest_func : ``Callable``
        Function object of the statistical test
    """
    for randtest_name in _randtests.__all__:
        randtest_func = getattr(_randtests, randtest_name)

        yield randtest_name, randtest_func


class TestNotFoundError(ValueError):
    """Error for when a statistical test is not found"""


columns = (
    TextColumn("[progress.description]{task.description}"),
    BarColumn(bar_width=54),  # i.e. progress is roughly 80 cols wide overall
    TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
    TimeRemainingColumn(),
)


@binary_check
def run_test(series: pd.Series, randtest_name, **kwargs) -> TestResult:
    """Run a statistical test on RNG output

    Parameters
    ----------
    series : ``Series``
        Output of the RNG being tested
    randtest_name : ``str``
        Name of statistical test
    **kwargs
        Keyword arguments to pass to statistical test

    Returns
    -------
    result : ``TestResult`` or ``MultiTestResult``
        Data containers of the test's result(s).

    Raises
    ------
    TestNotFoundError
        If `randtest_name` does not match any available statistical tests
    TestError
        Errors raised when running ``randtest_name``
    """
    try:
        func = getattr(_randtests, randtest_name)
    except AttributeError as e:
        raise TestNotFoundError() from e

    with Progress(*columns, console=console, transient=True) as progress:
        abbrv = f_randtest_abbreviations[randtest_name]
        task = progress.add_task(abbrv)

        try:
            result = func(series, ctx=(progress, task), **kwargs)

            color = "yellow" if result.failures else "green"
            print_randtest_name(randtest_name, color)
            console.print(result)

            return result

        except TestError as e:
            print_randtest_name(randtest_name, "red")
            print_error(e)

            raise e


@binary_check
def run_all_tests(series: pd.Series) -> Iterator[Tuple[str, TestResult, Exception]]:
    """Run all available statistical test on RNG output

    Yields
    ------
    randtest_name : ``str``
        Name of statistical test
    result : ``TestResult`` or ``MultiTestResult``
        Data containers of the test's result(s)
    exception : ``NotImplementedError`` or ``MinimumInputError``
        The exception raised when running ``randtest_name``, otherwise ``None``.

    Raises
    ------
    NonBinarySequenceError
        If series contains a sequence made of non-binary values
    """
    with Progress(*columns, console=console, transient=True) as progress:
        names, funcs = zip(*list_tests())
        tasks = []
        for name in names:
            abbrv = f_randtest_abbreviations[name]
            task = progress.add_task(abbrv, start=False)

            tasks.append(task)

        results = {}
        for name, func, task in zip(names, funcs, tasks):
            progress.start_task(task)

            try:
                result = func(series, ctx=(progress, task))

                color = "yellow" if result.failures else "green"

                print_randtest_name(name, color)
                console.print(result)

                yield name, result, None

                results[name] = result

            except TestError as e:
                progress.update(task, completed=True)

                print_randtest_name(name, "red")
                print_error(e)

                yield name, None, e

                results[name] = None

            console.print("")

    print_results_summary(results)


def print_results(results: Dict[str, BaseTestResult]):
    for name, result in results.items():
        color = "yellow" if result.failures else "green"

        print_randtest_name(name, color)
        console.print(result)
        console.print("")

    print_results_summary(results)


def print_results_summary(results: Dict[str, BaseTestResult]):
    size = get_terminal_size()
    ncols = min(size.columns, 80)

    rule = Rule("Test Results Summary", style="bright_blue")
    console.print(rule, width=ncols)

    unsummarisable_fnames = []

    table = Table(box=box.DOUBLE, caption=f"using a significance level of {SIGLEVEL}")
    table.add_column("Statistical Test", justify="eft")
    table.add_column("p-value", justify="left")
    table.add_column("Verdict", justify="left")
    for name, result in results.items():
        f_name = f_randtest_names[name]

        if isinstance(result, MultiTestResult):
            unsummarisable_fnames.append(f_name)
            continue

        if result:
            f_pvalue = str(round(result.p, 3))
            f_pvalue += "0" * (5 - len(f_pvalue))  # zero pad

            success = result.p >= SIGLEVEL
            verdict = "PASS" if success else "FAIL"
            colour = "green" if success else "red"
            f_verdict = Text(verdict, style=colour)
        else:
            f_pvalue = "-"
            f_verdict = Text("N/A", style="yellow")

        table.add_row(f_name, f_pvalue, f_verdict)

    if unsummarisable_fnames:
        warn_msg = "Multiple test results are currently not summarisable:\n"
        warn_msg += "\n".join(f"  • {f_name}" for f_name in unsummarisable_fnames)
        print_warning(warn_msg)

        console.print("")

    console.print(table)



from dataclasses import dataclass
from math import erfc
from math import log
from math import sqrt

import pandas as pd
from scipy.fft import fft

from coinflip._randtests.common.core import *
from coinflip._randtests.common.exceptions import NonBinarySequenceError
from coinflip._randtests.common.result import TestResult
from coinflip._randtests.common.result import make_testvars_list
from coinflip._randtests.common.typing import Float
from coinflip._randtests.common.typing import Integer

__all__ = ["spectral"]


class NonBinaryTruncatedSequenceError(NonBinarySequenceError):
    """Error if truncated sequence does not contain only 2 distinct values"""

    def __str__(self):
        return (
            "When truncated into an even-length, sequence contains only 1 distinct value\n"
            "i.e. the sequence was originally binary, but now isn't"
        )


@randtest()
def spectral(series, heads, tails, ctx):
    n = len(series)

    set_task_total(ctx, 4)

    failures = check_recommendations(ctx, {"n ≥ 1000": n >= 1000})

    if n % 2 != 0:
        series = series[:-1]
        if series.nunique() != 2:
            raise NonBinaryTruncatedSequenceError()

    threshold = sqrt(log(1 / 0.05) * n)
    nbelow_expect = 0.95 * n / 2

    advance_task(ctx)

    oscillations = series.map({heads: 1, tails: -1})
    fourier = pd.Series(fft(oscillations))  # fft returns a ndarray

    advance_task(ctx)

    half_fourier = fourier[: n // 2]
    peaks = half_fourier.abs()

    nbelow = sum(peaks < threshold)

    advance_task(ctx)

    diff = nbelow - nbelow_expect
    normdiff = diff / sqrt((n * 0.95 * 0.05) / 4)

    p = erfc(abs(normdiff) / sqrt(2))

    advance_task(ctx)

    return SpectralTestResult(
        heads, tails, failures, normdiff, p, nbelow_expect, nbelow, diff,
    )


@dataclass
class SpectralTestResult(TestResult):
    nbelow_expect: Float
    nbelow: Integer
    diff: Float

    def __post_init__(self):
        pass

    def _render(self):
        yield self._pretty_result("normalised diff")

        yield make_testvars_list(
            "npeaks above threshold",
            ("actual", self.nbelow),
            ("expected", self.nbelow_expect),
            ("diff", self.diff),
        )



__all__ = [
    "TestError",
    "TestNotImplementedError",
    "TestInputError",
    "NonBinarySequenceError",
]


class TestError(Exception):
    """Base class for test-related errors"""


class TestNotImplementedError(TestError, NotImplementedError):
    """Error if test is not implemented to handle valid parameters"""


class TestInputError(TestError, ValueError):
    """Error if test cannot handle (invalid) parameters"""


class NonBinarySequenceError(TestInputError):
    """Error if sequence does not contain only 2 distinct values"""

    def __str__(self):
        return "Sequence does not contain only 2 distinct values (i.e. binary)"



import pickle
from dataclasses import dataclass
from pathlib import Path
from typing import Dict

import pandas as pd
from jinja2 import Environment
from jinja2 import FileSystemLoader
from jinja2.exceptions import TemplateNotFound

from coinflip._randtests.common.result import BaseTestResult
from coinflip.cli.pprint import print_warning

__all__ = ["store_results", "load_results", "write_report_doc"]


@dataclass
class Report:
    series: pd.Series
    results: Dict[str, BaseTestResult]


def store_results(series: pd.Series, results: Dict[str, BaseTestResult], out: str):
    report = Report(series, results)

    pickle.dump(report, open(out, "wb"))


def load_results(results_path: Path) -> Report:
    report = pickle.load(open(results_path, "rb"))

    return report


templates_dir = Path(__file__).parents[3] / "templates"
templates_loader = FileSystemLoader(templates_dir)
templates_env = Environment(loader=templates_loader)


def write_report_doc(report: Report, out: Path):
    doc = templates_env.get_template("index.html")

    result_markups = []
    for randtest, result in report.results.items():
        template_file = f"randtests/{randtest}.html"
        try:
            template = templates_env.get_template(template_file)
            markup = template.render(result=result)
            result_markups.append(markup)
        except TemplateNotFound:
            print_warning(f"{template_file} not found")

    with open(out, "w") as f:
        report_html = doc.render(result_markups=result_markups)
        f.write(report_html)



import pickle
from numbers import Real
from typing import List

from hypothesis import strategies as st
from hypothesis.stateful import RuleBasedStateMachine
from hypothesis.stateful import initialize
from hypothesis.stateful import rule
from hypothesis.strategies import SearchStrategy
from pytest import raises

from coinflip.collections import Bins

from ..strategies import real


@st.composite
def intervals(draw) -> SearchStrategy[List[Real]]:
    n = draw(st.integers(min_value=2))

    x = draw(real())
    intervals = [x]
    for _ in range(n - 1):
        x = draw(real().filter(lambda x: x not in intervals))
        intervals.append(x)

    return intervals


class BinsStateMachine(RuleBasedStateMachine):
    @initialize(intervals=intervals())
    def init_bin(self, intervals):
        self.bins = Bins(intervals)
        print(intervals)
        print(self.bins)

    @rule(i=real(), n=real(allow_infinity=False))
    def increment(self, i, n):
        value_expect = self.bins[i] + n
        self.bins[i] += n
        assert self.bins[i] == value_expect

    @rule()
    def pickle(self):
        pickled_bins = pickle.dumps(self.bins)
        assert self.bins == pickle.loads(pickled_bins)


TestBinsStateMachine = BinsStateMachine.TestCase


def test_duplicate_intervals():
    with raises(ValueError):
        Bins([0, 1, 2, 2, 3])



import pickle

from hypothesis import given
from hypothesis import strategies as st
from pytest import mark

from coinflip._randtests.common.typing import *
from coinflip._randtests.cusum import CusumTestResult
from coinflip._randtests.entropy import ApproximateEntropyTestResult
from coinflip._randtests.excursions import RandomExcursionsMultiTestResult
from coinflip._randtests.excursions import RandomExcursionsVariantMultiTestResult
from coinflip._randtests.fourier import SpectralTestResult
from coinflip._randtests.frequency import FrequencyWithinBlockTestResult
from coinflip._randtests.frequency import MonobitTestResult
from coinflip._randtests.matrix import BinaryMatrixRankTestResult
from coinflip._randtests.runs import LongestRunsTestResult
from coinflip._randtests.runs import RunsTestResult
from coinflip._randtests.serial import SerialMultiTestResult
from coinflip._randtests.template import NonOverlappingTemplateMatchingMultiTestResult
from coinflip._randtests.template import OverlappingTemplateMatchingTestResult
from coinflip._randtests.universal import UniversalTestResult
from coinflip.collections import Bins


def test_bins_pickle():
    bins = Bins([0, 1, 2, 3, 4])
    bins[100] += 1

    pickled_bins = pickle.dumps(bins)

    bins2 = pickle.loads(pickled_bins)

    assert bins2 == bins


# TODO expand this
st.register_type_strategy(
    Face,
    st.one_of(st.booleans(), st.characters(), st.floats(), st.integers(), st.text()),
)


@mark.parametrize(
    "cls",
    [
        MonobitTestResult,
        FrequencyWithinBlockTestResult,
        RunsTestResult,
        LongestRunsTestResult,
        BinaryMatrixRankTestResult,
        SpectralTestResult,
        NonOverlappingTemplateMatchingMultiTestResult,
        OverlappingTemplateMatchingTestResult,
        UniversalTestResult,
        SerialMultiTestResult,
        ApproximateEntropyTestResult,
        CusumTestResult,
        RandomExcursionsMultiTestResult,
        RandomExcursionsVariantMultiTestResult,
    ],
)
@given(st.data())
def test_results(cls, data):
    data.draw(st.builds(cls))



from collections import defaultdict
from dataclasses import dataclass
from math import ceil
from math import erfc
from math import floor
from math import isclose
from math import log
from math import log2
from math import sqrt
from typing import DefaultDict
from typing import List
from typing import NamedTuple
from typing import Tuple

from coinflip._randtests.common.collections import FloorDict
from coinflip._randtests.common.core import *
from coinflip._randtests.common.exceptions import TestNotImplementedError
from coinflip._randtests.common.result import TestResult
from coinflip._randtests.common.testutils import rawblocks
from coinflip._randtests.common.typing import Face
from coinflip._randtests.common.typing import Integer

__all__ = ["maurers_universal"]


class Dist(NamedTuple):
    mean: float
    variance: float


# Values taken from "A Handbook of Applied Cryptography"
#     See section 5.4.5, "Maurer's universal statistical test", p. 184
# TODO generate this dynamically
blocksize_dists = {
    1: Dist(0.7326495, 0.690),
    2: Dist(1.5374383, 1.338),
    3: Dist(2.4016068, 1.901),
    4: Dist(3.3112247, 2.358),
    5: Dist(4.2534266, 2.705),
    6: Dist(5.2177052, 2.954),
    7: Dist(6.1962507, 3.125),
    8: Dist(7.1836656, 3.238),
    9: Dist(8.1764248, 3.311),
    10: Dist(9.1723243, 3.356),
    11: Dist(10.170032, 3.384),
    12: Dist(11.168765, 3.401),
    13: Dist(12.168070, 3.410),
    14: Dist(13.167693, 3.416),
    15: Dist(14.167488, 3.419),
    16: Dist(15.167379, 3.421),
}


class DefaultParams(NamedTuple):
    blocksize: Integer
    init_nblocks: Integer


# Values taken from "A Statistical Test Suite for Random and Pseudorandom Number
#                    Generators for Cryptographic Applications"
#     See section 2.9.7, "Input Size Recommendation", p. 45
n_defaults = FloorDict(
    {
        387840: DefaultParams(6, 640),
        904960: DefaultParams(7, 1280),
        2068480: DefaultParams(8, 2560),
        4654080: DefaultParams(9, 5120),
        10342400: DefaultParams(10, 10240),
        22753280: DefaultParams(11, 20480),
        49643520: DefaultParams(12, 40960),
        107560960: DefaultParams(13, 81920),
        231669760: DefaultParams(14, 163840),
        496435200: DefaultParams(15, 327680),
        1059061760: DefaultParams(16, 655360),
    }
)


@randtest(min_n=4)
def maurers_universal(series, heads, tails, ctx, blocksize=None, init_nblocks=None):
    if blocksize and blocksize > 16:
        # TODO review this policy
        raise TestNotImplementedError(
            "Test implementation cannot handle blocksize over 16"
        )

    n = len(series)

    if not blocksize or not init_nblocks:
        try:
            blocksize, init_nblocks = n_defaults[n]
        except KeyError:
            blocksize = min(max(ceil(log(n)), 2), 16)  # largest blocksize_dists key
            nblocks = n // blocksize
            init_nblocks = max(nblocks // 100, 1)

    init_n = init_nblocks * blocksize
    init_series, segment_series = series[:init_n], series[init_n:]
    segment_nblocks = (n - init_n) // blocksize

    mean_expect, variance = blocksize_dists[blocksize]

    set_task_total(ctx, init_nblocks + segment_nblocks + 2)

    failures = check_recommendations(
        ctx,
        {
            "n ≥ 387840": n >= 387840,
            "6 ≤ blocksize ≤ 16": 6 <= blocksize <= 16,
            "init_nblocks ≈ 10 * 2 ** segment_nblocks": intclose(
                init_nblocks, 10 * 2 ** segment_nblocks
            ),
            "segment_nblocks ≈ ⌈n / blocksize⌉ - init_nblocks": isclose(
                segment_nblocks, floor(n / blocksize) - init_nblocks
            ),
        },
    )

    permutation_last_init_pos = defaultdict(int)

    init_blocks = rawblocks(init_series, blocksize)
    for pos, permutation in enumerate(init_blocks, 1):
        permutation_last_init_pos[permutation] = pos

        advance_task(ctx)

    segment_blocks = rawblocks(segment_series, blocksize)
    permutation_positions = defaultdict(list)
    segment_firstpos = init_nblocks + 1
    for pos, permutation in enumerate(segment_blocks, segment_firstpos):
        permutation_positions[permutation].append(pos)

        advance_task(ctx)

    distances_total = 0
    for permutation, positions in permutation_positions.items():
        last_occurence = permutation_last_init_pos[permutation]
        for pos in positions:
            distance = pos - last_occurence
            distances_total += log2(distance)

            last_occurence = pos

    advance_task(ctx)

    statistic = distances_total / segment_nblocks

    normdiff = abs((statistic - mean_expect) / (sqrt(2 * variance)))
    p = erfc(normdiff)

    advance_task(ctx)

    return UniversalTestResult(
        heads,
        tails,
        failures,
        statistic,
        p,
        blocksize,
        init_nblocks,
        segment_nblocks,
        permutation_last_init_pos,
        permutation_positions,
    )


@dataclass
class UniversalTestResult(TestResult):
    blocksize: Integer
    init_nblocks: Integer
    segment_nblocks: Integer
    permutation_last_init_pos: DefaultDict[Tuple[Face, ...], Integer]
    permutation_positions: DefaultDict[Tuple[Face, ...], List[Integer]]

    def _render(self):
        yield self._pretty_result("log2 distances")

        yield TestResult._pretty_inputs(
            ("init nblocks", self.init_nblocks),
            ("segment nblocks", self.segment_nblocks),
        )

        # TODO maybe use this table for a verbose option or something (it's huge)
        # table = make_testvars_table("permutation", "init pos", "test positions", justify=False)
        # for permutation, positions in self.permutation_positions.items():
        #     f_permutation = pretty_subseq(permutation, self.heads, self.tails)

        #     init_pos = self.permutation_last_init_pos[permutation]
        #     f_init_pos = str(init_pos)

        #     f_positions = ", ".join(str(pos) for pos in positions)

        #     table.add_row(f_permutation, f_init_pos, f_positions)


def intclose(int1, int2, abs_tol=5):
    return abs(int1 - int2) <= abs_tol



from functools import lru_cache
from typing import Iterable
from typing import Tuple

from rich.style import Style
from rich.table import Table
from rich.text import Text

__all__ = ["make_warning", "determine_rep", "pretty_subseq"]


warn_text = Text("WARN", style="yellow dim")


def make_warning(msg: str) -> Table:
    f_msg = Text(msg, style="dim")

    grid = Table.grid(padding=(1))
    grid.add_row(warn_text, f_msg)

    return grid


@lru_cache()
def determine_rep(heads, tails) -> Tuple[str, str]:
    """Determine single-character representations of each binary value

    Parameters
    ----------
    heads: ``Face``
        The ``1`` abstraction
    tails: ``Face``
        The ``0`` abstraction

    Returns
    -------
    heads_rep : ``str``
        Character representation of the ``heads``
    tails_rep : ``str``
        Character representation of the ``tails``
    """
    heads_rep = str(heads)[0]
    tails_rep = str(tails)[0]
    if heads_rep == tails_rep:
        heads_rep = "1"
        tails_rep = "0"

    return heads_rep, tails_rep


bright = Style(bold=True, dim=False)


def pretty_subseq(subseq: Iterable, heads, tails) -> Text:
    """Produce a one-line pretty representation of a subsequence

    Parameters
    ----------
    subseq : ``List``
        Subsequence to represent
    heads : ``Face``
        One of the two values in ``subseq``
    tails : ``Face``
        Value in ``subseq`` which is not ``heads``

    Returns
    -------
    subseq_rep: ``Text``
        Pretty representation of ``subseq``
    """
    heads_rep, tails_rep = determine_rep(heads, tails)

    subseq_rep = ""
    for element in subseq:
        if element == heads:
            subseq_rep += heads_rep
        else:
            subseq_rep += tails_rep

    return Text(subseq_rep, style=bright)



from dataclasses import dataclass
from math import floor
from math import sqrt

import numpy as np
from scipy.stats import norm

from coinflip._randtests.common.core import *
from coinflip._randtests.common.result import TestResult
from coinflip._randtests.common.typing import Bool

__all__ = ["cusum"]


@randtest()
def cusum(series, heads, tails, ctx, reverse=False):
    n = len(series)

    set_task_total(ctx, 3)

    failures = check_recommendations(ctx, {"n ≥ 100": n >= 100})

    oscillations = series.map({heads: 1, tails: -1})

    advance_task(ctx)

    if reverse:
        oscillations = oscillations[::-1]

    cusums = oscillations.cumsum()
    abs_cusums = cusums.abs()
    max_cusum = abs_cusums.nlargest(1).iloc[0]

    advance_task(ctx)

    # TODO this all can be done more elegantly
    start1 = floor((-n / max_cusum + 1) / 4)
    start2 = floor((-n / max_cusum - 3) / 4)
    stop = floor((n / max_cusum - 1) / 4) + 1
    p = (
        1
        - sum(
            norm.cdf((4 * k + 1) * max_cusum / sqrt(n))
            - norm.cdf((4 * k - 1) * max_cusum / sqrt(n))
            for k in np.arange(start1, stop, 1)
        )
        + sum(
            norm.cdf((4 * k + 3) * max_cusum / sqrt(n))
            - norm.cdf((4 * k + 1) * max_cusum / sqrt(n))
            for k in np.arange(start2, stop, 1)
        )
    )

    advance_task(ctx)

    return CusumTestResult(heads, tails, failures, max_cusum, p, reverse)


@dataclass
class CusumTestResult(TestResult):
    reverse: Bool

    def _render(self):
        yield self._pretty_result("max cusum")



from coinflip._randtests.complexity import linear_complexity
from coinflip._randtests.cusum import cusum
from coinflip._randtests.entropy import approximate_entropy
from coinflip._randtests.excursions import random_excursions
from coinflip._randtests.excursions import random_excursions_variant
from coinflip._randtests.fourier import spectral
from coinflip._randtests.frequency import frequency_within_block
from coinflip._randtests.frequency import monobit
from coinflip._randtests.matrix import binary_matrix_rank
from coinflip._randtests.runs import longest_runs
from coinflip._randtests.runs import runs
from coinflip._randtests.serial import serial
from coinflip._randtests.template import non_overlapping_template_matching
from coinflip._randtests.template import overlapping_template_matching
from coinflip._randtests.universal import maurers_universal

__all__ = [
    "monobit",
    "frequency_within_block",
    "runs",
    "longest_runs",
    "binary_matrix_rank",
    "spectral",
    "non_overlapping_template_matching",
    "overlapping_template_matching",
    "maurers_universal",
    "linear_complexity",
    "serial",
    "approximate_entropy",
    "cusum",
    "random_excursions",
    "random_excursions_variant",
]



"""Assert test results from examples on our implementations"""
from math import isclose
from pathlib import Path
from typing import Any
from typing import Dict
from typing import Iterator
from typing import List
from typing import NamedTuple
from typing import Tuple
from typing import Union

from pytest import mark
from pytest import param
from typing_extensions import Literal

__all__ = [
    "example_fields",
    "multi_example_fields",
    "sub_example_fields",
    "examples",
    "multi_examples",
    "sub_examples",
    "assert_statistic",
    "assert_p",
    "assert_statistics",
    "assert_pvalues",
]

e_path = Path(__file__).parent / "e_expansion.txt"


def e_expansion(n=1000000) -> Iterator[int]:
    """Generates bits of e

    Note
    ----
    Uses the same bit expansion that's included in SP800-22's `sts`
    """

    def genbits():
        with open(e_path) as f:
            for line in f:
                for x in line:
                    if x == "0" or x == "1":
                        bit = int(x)

                        yield bit

    e = genbits()
    for _ in range(n):
        yield next(e)


class Example(NamedTuple):
    """Container for a SP800-22 example"""

    randtest: str
    bits: List[Literal[0, 1]]
    statistic_expect: Union[int, float]
    p_expect: float
    kwargs: Dict[str, Any] = {}
    xfail: bool = False


class MultiExample(NamedTuple):
    """Container for a SP800-22 example with multiple results"""

    randtest: str
    bits: List[Literal[0, 1]]
    expected_statistics: List[Union[int, float]]
    expected_pvalues: List[float]
    kwargs: Dict[str, Any] = {}
    xfail: bool = False


class SubExample(NamedTuple):
    """Container for a single SP800-22 example of a test with multiple results"""

    randtest: str
    key: Any
    bits: List[Literal[0, 1]]
    statistic_expect: Union[int, float]
    p_expect: float
    kwargs: Dict[str, Any] = {}
    xfail: bool = False


example_fields = ["randtest", "bits", "statistic_expect", "p_expect", "kwargs"]
multi_example_fields = [
    "randtest",
    "bits",
    "expected_statistics",
    "expected_pvalues",
    "kwargs",
]
sub_example_fields = [
    "randtest",
    "key",
    "bits",
    "statistic_expect",
    "p_expect",
    "kwargs",
]


# fmt: off
_examples = [
    Example(
        randtest="monobit",

        bits=[1, 0, 1, 1, 0, 1, 0, 1, 0, 1],

        statistic_expect=.632455532,
        p_expect=0.527089,
    ),
    Example(
        randtest="monobit",

        bits=[
            1, 1, 0, 0, 1, 0, 0, 1,
            0, 0, 0, 0, 1, 1, 1, 1,
            1, 1, 0, 1, 1, 0, 1, 0,
            1, 0, 1, 0, 0, 0, 1, 0,
            0, 0, 1, 0, 0, 0, 0, 1,
            0, 1, 1, 0, 1, 0, 0, 0,
            1, 1, 0, 0, 0, 0, 1, 0,
            0, 0, 1, 1, 0, 1, 0, 0,
            1, 1, 0, 0, 0, 1, 0, 0,
            1, 1, 0, 0, 0, 1, 1, 0,
            0, 1, 1, 0, 0, 0, 1, 0,
            1, 0, 0, 0, 1, 0, 1, 1,
            1, 0, 0, 0,
        ],
        statistic_expect=1.6,
        p_expect=0.109599,
    ),
    Example(
        randtest="frequency_within_block",

        bits=[
            1, 1, 0, 0, 1, 0, 0, 1, 0, 0,
            0, 0, 1, 1, 1, 1, 1, 1, 0, 1,
            1, 0, 1, 0, 1, 0, 1, 0, 0, 0,
            1, 0, 0, 0, 1, 0, 0, 0, 0, 1,
            0, 1, 1, 0, 1, 0, 0, 0, 1, 1,
            0, 0, 0, 0, 1, 0, 0, 0, 1, 1,
            0, 1, 0, 0, 1, 1, 0, 0, 0, 1,
            0, 0, 1, 1, 0, 0, 0, 1, 1, 0,
            0, 1, 1, 0, 0, 0, 1, 0, 1, 0,
            0, 0, 1, 0, 1, 1, 1, 0, 0, 0,
        ],
        kwargs={
            "blocksize": 10
        },

        statistic_expect=7.2,
        p_expect=0.706438,
    ),
    Example(
        randtest="runs",

        bits=[
            1, 0, 0, 1, 1, 0, 1, 0,
            1, 1
        ],

        statistic_expect=7,
        p_expect=0.147232,
    ),
    Example(
        randtest="longest_runs",

        bits=[
            1, 1, 0, 0, 1, 1, 0, 0,
            0, 0, 0, 1, 0, 1, 0, 1,
            0, 1, 1, 0, 1, 1, 0, 0,
            0, 1, 0, 0, 1, 1, 0, 0,
            1, 1, 1, 0, 0, 0, 0, 0,
            0, 0, 0, 0, 0, 0, 1, 0,
            0, 1, 0, 0, 1, 1, 0, 1,
            0, 1, 0, 1, 0, 0, 0, 1,
            0, 0, 0, 1, 0, 0, 1, 1,
            1, 1, 0, 1, 0, 1, 1, 0,
            1, 0, 0, 0, 0, 0, 0, 0,
            1, 1, 0, 1, 0, 1, 1, 1,
            1, 1, 0, 0, 1, 1, 0, 0,
            1, 1, 1, 0, 0, 1, 1, 0,
            1, 1, 0, 1, 1, 0, 0, 0,
            1, 0, 1, 1, 0, 0, 1, 0,
        ],


        statistic_expect=4.882605,
        p_expect=0.180609,
    ),
    Example(
        randtest="binary_matrix_rank",

        bits=[
            0, 1, 0, 1, 1, 0, 0, 1,
            0, 0, 1, 0, 1, 0, 1, 0,
            1, 1, 0, 1,
        ],
        kwargs={
            "matrix_dimen": (3, 3),
        },

        statistic_expect=0.596953,
        p_expect=0.741948,
    ),
    Example(
        randtest="binary_matrix_rank",

        bits=list(e_expansion(n=100000)),
        kwargs={
            "matrix_dimen": (32, 32),
        },

        statistic_expect=1.2619656,
        p_expect=0.532069,
    ),
    Example(
        # FAILING scipys fft produces slightly diff transformations to SP800-22's sts
        #         TODO  - check if examples succeed using SP800-22's fourier transforming
        #               - performance metrics on both solutions
        randtest="spectral",

        bits=[1, 0, 0, 1, 0, 1, 0, 0, 1, 1],

        statistic_expect=-2.176429,
        p_expect=0.029523,

        xfail=True,
    ),
    Example(
        randtest="spectral",

        bits=[
            1, 1, 0, 0, 1, 0, 0, 1,
            0, 0, 0, 0, 1, 1, 1, 1,
            1, 1, 0, 1, 1, 0, 1, 0,
            1, 0, 1, 0, 0, 0, 1, 0,
            0, 0, 1, 0, 0, 0, 0, 1,
            0, 1, 1, 0, 1, 0, 0, 0,
            1, 1, 0, 0, 0, 0, 1, 0,
            0, 0, 1, 1, 0, 1, 0, 0,
            1, 1, 0, 0, 0, 1, 0, 0,
            1, 1, 0, 0, 0, 1, 1, 0,
            0, 1, 1, 0, 0, 0, 1, 0,
            1, 0, 0, 0, 1, 0, 1, 1,
            1, 0, 0, 0,
        ],

        statistic_expect=-1.376494,
        p_expect=0.168669,

        xfail=True,
    ),
    Example(
        # FAILING p off by ~0.07 if gammaincc(df/2, statistic/2) and df=2
        randtest="overlapping_template_matching",

        bits=[
            1, 0, 1, 1, 1, 0, 1, 1, 1, 1,
            0, 0, 1, 0, 1, 1, 0, 1, 1, 0,  # Modifed 2nd block of SP800-22 example
            0, 1, 1, 1, 0, 0, 1, 0, 1, 1,  # originally had 1 match
            1, 0, 1, 1, 1, 1, 1, 0, 0, 0,  # now has 2 matches, as expected
            0, 1, 0, 1, 1, 0, 1, 0, 0, 1,
        ],
        kwargs={
            "template_size": 2,
            "blocksize": 10,  # nblocks=5
            "df": 2,
        },

        statistic_expect=3.167729,
        p_expect=0.274932,

        xfail=True
    ),
    Example(
        randtest="overlapping_template_matching",

        bits=list(e_expansion()),
        kwargs={
            "template_size": 9,
            "blocksize": 1032,  # nblocks=968
        },

        statistic_expect=8.965859,
        p_expect=0.110434
    ),
    Example(
        randtest="maurers_universal",

        bits=[
            0, 1, 0, 1, 1, 0, 1, 0,
            0, 1, 1, 1, 0, 1, 0, 1,
            0, 1, 1, 1,
        ],
        kwargs={
            "blocksize": 2,
            "init_nblocks": 4,
        },

        statistic_expect=1.1949875,
        p_expect=0.767189,
    ),
    Example(
        randtest="linear_complexity",

        bits=list(e_expansion()),
        kwargs={
            "blocksize": 1000
        },

        statistic_expect=2.700348,
        p_expect=0.845406,
    ),
    Example(
        randtest="approximate_entropy",

        bits=[0, 1, 0, 0, 1, 1, 0, 1, 0, 1],
        kwargs={
            "blocksize": 3
        },

        statistic_expect=10.043859999999999,  # SP800-22 erroneously had 0.502193
        p_expect=0.261961,
    ),
    Example(
        randtest="approximate_entropy",

        bits=[
            1, 1, 0, 0, 1, 0, 0, 1,
            0, 0, 0, 0, 1, 1, 1, 1,
            1, 1, 0, 1, 1, 0, 1, 0,
            1, 0, 1, 0, 0, 0, 1, 0,
            0, 0, 1, 0, 0, 0, 0, 1,
            0, 1, 1, 0, 1, 0, 0, 0,
            1, 1, 0, 0, 0, 0, 1, 0,
            0, 0, 1, 1, 0, 1, 0, 0,
            1, 1, 0, 0, 0, 1, 0, 0,
            1, 1, 0, 0, 0, 1, 1, 0,
            0, 1, 1, 0, 0, 0, 1, 0,
            1, 0, 0, 0, 1, 0, 1, 1,
            1, 0, 0, 0,
        ],
        kwargs={
            "blocksize": 2
        },

        statistic_expect=5.550792,
        p_expect=0.235301,
    ),
    Example(
        randtest="cusum",

        bits=[1, 0, 1, 1, 0, 1, 0, 1, 1, 1],

        statistic_expect=4,
        p_expect=0.4116588,
    ),
    Example(
        randtest="cusum",

        bits=[
            1, 1, 0, 0, 1, 0, 0, 1,
            0, 0, 0, 0, 1, 1, 1, 1,
            1, 1, 0, 1, 1, 0, 1, 0,
            1, 0, 1, 0, 0, 0, 1, 0,
            0, 0, 1, 0, 0, 0, 0, 1,
            0, 1, 1, 0, 1, 0, 0, 0,
            1, 1, 0, 0, 0, 0, 1, 0,
            0, 0, 1, 1, 0, 1, 0, 0,
            1, 1, 0, 0, 0, 1, 0, 0,
            1, 1, 0, 0, 0, 1, 1, 0,
            0, 1, 1, 0, 0, 0, 1, 0,
            1, 0, 0, 0, 1, 0, 1, 1,
            1, 0, 0, 0,
        ],

        statistic_expect=16,
        p_expect=0.219194,
    ),
    Example(
        randtest="cusum",

        bits=[
            1, 1, 0, 0, 1, 0, 0, 1,
            0, 0, 0, 0, 1, 1, 1, 1,
            1, 1, 0, 1, 1, 0, 1, 0,
            1, 0, 1, 0, 0, 0, 1, 0,
            0, 0, 1, 0, 0, 0, 0, 1,
            0, 1, 1, 0, 1, 0, 0, 0,
            1, 1, 0, 0, 0, 0, 1, 0,
            0, 0, 1, 1, 0, 1, 0, 0,
            1, 1, 0, 0, 0, 1, 0, 0,
            1, 1, 0, 0, 0, 1, 1, 0,
            0, 1, 1, 0, 0, 0, 1, 0,
            1, 0, 0, 0, 1, 0, 1, 1,
            1, 0, 0, 0,
        ],
        kwargs={
            "reverse": True
        },

        statistic_expect=19,
        p_expect=0.114866,
    ),
]

_multi_examples = [
    MultiExample(
        # FAILING - SP800-22's result is not replicated by sts
        #         - sts and dj result matches our own
        randtest="serial",

        bits=[0, 0, 1, 1, 0, 1, 1, 1, 0, 1],
        kwargs={
            "blocksize": 3
        },

        expected_statistics=[1.6, 0.8],
        expected_pvalues=[0.9057, 0.8805],

        xfail=True
    ),
    MultiExample(
        randtest="serial",

        bits=list(e_expansion()),
        kwargs={
            "blocksize": 2
        },

        expected_statistics=[0.339764, 0.336400],
        expected_pvalues=[0.843764, 0.561915],
    ),
    MultiExample(
        # FAILING - SP800-22's result is not replicated by sts
        #         - sts result matches our own
        #         - dj near to our own
        randtest="random_excursions",

        bits=list(e_expansion()),

        expected_statistics=[
            3.835698,
            7.318707,
            7.861927,
            15.692617,
            2.485906,
            5.429381,
            2.404171,
            2.393928,
        ],
        expected_pvalues=[
            0.573306,
            0.197996,
            0.164011,
            0.007779,
            0.778616,
            0.365752,
            0.790853,
            0.792378,
        ],

        xfail=True
    ),
    MultiExample(
        randtest="random_excursions_variant",

        bits=list(e_expansion()),

        expected_statistics=[
            1450,
            1435,
            1380,
            1366,
            1412,
            1475,
            1480,
            1468,
            1502,
            1409,
            1369,
            1396,
            1479,
            1599,
            1628,
            1619,
            1620,
            1610,
        ],
        expected_pvalues=[
            0.858946,
            0.794755,
            0.576249,
            0.493417,
            0.633873,
            0.917283,
            0.934708,
            0.816012,
            0.826009,
            0.137861,
            0.200642,
            0.441254,
            0.939291,
            0.505683,
            0.445935,
            0.512207,
            0.538635,
            0.593930,
        ]
    )
]

_sub_examples = [
    SubExample(
        randtest="non_overlapping_template_matching",

        key=(0, 0, 1),

        bits=[
            1, 0, 1, 0, 0, 1, 0, 0,
            1, 0, 1, 1, 1, 0, 0, 1,
            0, 1, 1, 0,
        ],
        kwargs={
            "template_size": 3,
            "blocksize": 10,  # nblocks=2
        },

        statistic_expect=2.133333,
        p_expect=0.344154,
    ),
    SubExample(
        randtest="random_excursions",

        key=1,

        bits=[0, 1, 1, 0, 1, 1, 0, 1, 0, 1],

        statistic_expect=4.333033,
        p_expect=0.502529,
    ),
    SubExample(
        randtest="random_excursions_variant",

        key=1,

        bits=[0, 1, 1, 0, 1, 1, 0, 1, 0, 1],

        statistic_expect=4,
        p_expect=0.683091,
    ),
]
# fmt: on


# TODO make a type hint for all examples e.g. BaseExample
def wrap_examples(examples_list: List) -> List[Union[Tuple, "ParameterSet"]]:  # noqa
    argvalues = []
    for example in examples_list:
        *_argvalue, xfail = example
        marks = []
        if len(example.bits) > 1000:
            marks.append(mark.slow)
        if xfail:
            marks.append(mark.xfail)

        if marks:
            argvalue = param(*_argvalue, marks=marks)
        else:
            argvalue = _argvalue

        argvalues.append(argvalue)

    return argvalues


examples = wrap_examples(_examples)
multi_examples = wrap_examples(_multi_examples)
sub_examples = wrap_examples(_sub_examples)


def assert_statistic(statistic, statistic_expect):
    if isinstance(statistic, int):
        assert (
            statistic == statistic_expect
        ), f"statistic {statistic} != {statistic_expect}"
    else:
        assert isclose(
            statistic, statistic_expect, rel_tol=0.05
        ), f"statistic {round(statistic, 1)} != {round(statistic_expect, 1)}"


def assert_p(p, p_expect):
    assert isclose(
        p, p_expect, abs_tol=0.005
    ), f"p-value {round(p, 3)} !≈ {round(p_expect, 3)}"


def assert_statistics(statistics, expected_statistics):
    for statistic, statistic_expect in zip(statistics, expected_statistics):
        assert_statistic(statistic, statistic_expect)


def assert_pvalues(pvalues, expected_pvalues):
    for p, p_expect in zip(pvalues, expected_pvalues):
        assert_p(p, p_expect)



from copy import copy
from dataclasses import dataclass
from math import floor
from math import sqrt
from typing import Dict
from typing import List
from typing import Sequence

from scipy.stats import chisquare
from typing_extensions import Literal

from coinflip._randtests.common.collections import Bins
from coinflip._randtests.common.core import *
from coinflip._randtests.common.result import TestResult
from coinflip._randtests.common.result import make_chisquare_table
from coinflip._randtests.common.result import smartround
from coinflip._randtests.common.testutils import rawblocks
from coinflip._randtests.common.typing import Float
from coinflip._randtests.common.typing import Integer

__all__ = ["linear_complexity"]


# TODO - remove hardcoded degrees of freedom
#      - dynamically calculate these probabilities
df = 6
probabilities = [
    0.010417,
    0.03125,
    0.125,
    0.5,
    0.25,
    0.0625,
    0.020833,
]


@randtest()
def linear_complexity(series, heads, tails, ctx, blocksize=None):
    n = len(series)

    if not blocksize:
        for blocksize in [1000, 2500, 500, 5000]:
            nblocks = n // blocksize
            if 500 <= blocksize <= 5000 and nblocks >= 200:
                break
        else:
            blocksize = max(floor(sqrt(n)), 2)

    nblocks = n // blocksize

    set_task_total(ctx, nblocks + 3)

    failures = check_recommendations(
        ctx,
        {
            "n ≥ 1000000": n >= 1000000,
            "500 ≤ blocksize ≤ 5000": 500 <= blocksize <= 5000,
            "nblocks ≥ 200": nblocks >= 200,
        },
    )

    binary = series.map({heads: 1, tails: 0})

    advance_task(ctx)

    mean_expect = (
        blocksize / 2
        + (9 + (-(1 ** (blocksize + 1)))) / 36
        - (blocksize / 3 + 2 / 9) / 2 ** blocksize
    )
    expected_bincounts = [nblocks * prob for prob in probabilities]

    advance_task(ctx)

    variance_bins = Bins([-3, -2, -1, 0, 1, 2, 3])
    for block_tup in rawblocks(binary, blocksize):
        linear_complexity = berlekamp_massey(block_tup)
        variance = (-1) ** blocksize * (linear_complexity - mean_expect) + 2 / 9
        variance_bins[variance] += 1

        advance_task(ctx)

    statistic, p = chisquare(list(variance_bins.values()), expected_bincounts)

    advance_task(ctx)

    return LinearComplexityTestResult(
        heads,
        tails,
        failures,
        statistic,
        p,
        blocksize,
        mean_expect,
        expected_bincounts,
        variance_bins,
    )


@dataclass
class LinearComplexityTestResult(TestResult):
    blocksize: Integer
    mean_expect: Float
    expected_bincounts: List[Float]
    variance_bins: Dict[Float, Integer]

    def _render(self):
        yield self._pretty_result("chi-square")

        f_mean_expect = smartround(self.mean_expect)

        table = make_chisquare_table(
            "linear complexity deviations per block",
            "variance",
            self.variance_bins.keys(),
            self.expected_bincounts,
            self.variance_bins.values(),
            caption=f"from expected mean {f_mean_expect}",
        )
        yield table


def berlekamp_massey(sequence: Sequence[Literal[0, 1]]) -> int:
    """Finds the shortest LSFR in a sequence"""
    n = len(sequence)

    error_locator = [0 for _ in range(n)]
    error_locator[0] = 1

    error_locator_prev = copy(error_locator)

    min_size = 0  # of the LSFR
    nloops = -1  # since error_locator_prev and min_size were updated

    for i, seq_bit in enumerate(sequence):
        discrepancy = seq_bit
        seq_window = reversed(sequence[i - min_size : i])
        errloc_window = error_locator[1 : min_size + 1]
        for seq_bit, errloc_bit in zip(seq_window, errloc_window):
            product = seq_bit & errloc_bit
            discrepancy = discrepancy ^ product

        if discrepancy:
            error_locator_temp = copy(error_locator)

            recalc_bits = []
            for errloc_bit, prev_bit in zip(
                error_locator[i - nloops : n], error_locator_prev
            ):
                bit = errloc_bit ^ prev_bit
                recalc_bits.append(bit)
            error_locator[i - nloops : n] = recalc_bits

            if min_size <= i / 2:
                min_size = i + 1 - min_size
                nloops = i
                error_locator_prev = error_locator_temp

    return min_size



import webbrowser
from tempfile import NamedTemporaryFile

from click.testing import CliRunner
from hypothesis import HealthCheck
from hypothesis import settings
from hypothesis.stateful import Bundle
from hypothesis.stateful import RuleBasedStateMachine
from hypothesis.stateful import rule

from coinflip.cli import commands
from coinflip.cli import console

from .strategies import mixedbits


class CliStateMachine(RuleBasedStateMachine):
    def __init__(self):
        super(CliStateMachine, self).__init__()
        self.runner = CliRunner()

        console.print = noop
        webbrowser.open = noop

    randtest_results = Bundle("randtest_results")
    reports = Bundle("reports")

    @rule()
    def main(self):
        result = self.runner.invoke(commands.main, [])
        assert_success(result)

    @rule()
    def example_run(self):
        result = self.runner.invoke(commands.example_run, [])
        assert_success(result)

    @rule(target=randtest_results, sequence=mixedbits())
    def run(self, sequence):
        data = NamedTemporaryFile()
        out = NamedTemporaryFile()

        with data as f:
            for x in sequence:
                x_bin = str(x).encode("utf-8")
                line = x_bin + b"\n"
                f.write(line)

            f.seek(0)
            result = self.runner.invoke(commands.run, [f.name, out.name])

        assert_success(result)

        return out.name

    @rule(path=randtest_results)
    def read(self, path):
        result = self.runner.invoke(commands.read, [path])
        assert_success(result)

    @rule(target=reports, path=randtest_results)
    def report(self, path):
        out = NamedTemporaryFile()

        result = self.runner.invoke(commands.report, [path, out.name])
        assert_success(result)

        return out.name


TestCliStateMachine = CliStateMachine.TestCase  # top-level TestCase picked up by pytest
TestCliStateMachine.settings = settings(
    suppress_health_check=[HealthCheck.data_too_large, HealthCheck.too_slow],
    deadline=None,
)


def noop(*args, **kwargs):
    return None


def assert_success(result):
    assert result.exit_code == 0, result.stderr if result.stderr_bytes else str(result)



from shutil import get_terminal_size

from rich.console import RenderGroup
from rich.console import render_group
from rich.rule import Rule
from rich.style import Style
from rich.text import Text

from coinflip._randtests.common.core import infer_faces
from coinflip._randtests.common.pprint import make_warning
from coinflip._randtests.common.pprint import pretty_subseq
from coinflip._randtests.common.testutils import rawblocks
from coinflip.cli import console

__all__ = ["print_warning", "print_error", "print_series"]


err_text = Text("ERR!", style="red")


def print_error(e: Exception):
    """Pretty print exceptions"""
    text = Text(style="bright")
    text.append(err_text)
    text.append(f" {e}")

    console.print(text)


def print_warning(msg: str):
    f_warning = make_warning(msg)

    console.print(f_warning)


dim = Style(dim=True)


# TODO descriptions of the series e.g. length
def print_series(series):
    """Pretty print series that contain binary data"""
    size = get_terminal_size()
    ncols = min(size.columns, 80)

    rule = Rule("Sequence To Test", style="bright_blue")
    console.print(rule, width=ncols)

    f_series = pretty_sequence(series, ncols)
    console.print(f_series)


@render_group(fit=True)
def pretty_sequence(series, ncols) -> RenderGroup:
    """Produce a multi-line representation of a sequence

    Parameters
    ----------
    series : ``Series``
        Sequence to represent
    ncols : ``Int``
        Maximum number of characters to use per line

    Returns
    -------
    series_rep : ``RenderGroup``
        Pretty represented of a sequence
    """
    values = series.unique()
    heads, tails = infer_faces(tuple(values))

    gap = 2

    outer_w = ncols - 2 * gap
    inner_w = outer_w - 2 * gap

    pad = Text("".join(" " for _ in range(gap)))

    l_border = Text("  | ", style=dim)
    r_border = Text(" |  ", style=dim)

    l_arrow = Text(" <  ", style=dim)
    r_arrow = Text("  > ", style=dim)

    def make_hline(width) -> Text:
        return Text("+" + "".join("-" for _ in range(width - 2)) + "+", style=dim)

    def pretty_row(row) -> Text:
        return pretty_subseq(row, heads, tails)

    n = len(series)
    if n <= inner_w:
        border = pad + make_hline(n + 4) + pad

        yield border
        yield l_border + pretty_row(series.array) + r_border
        yield border

    else:
        border = pad + make_hline(outer_w) + pad
        rows = list(rawblocks(series, blocksize=inner_w, truncate=False))

        yield border
        yield l_border + pretty_row(rows[0]) + r_arrow
        yield border

        nrows = len(rows)
        if nrows <= 12:
            for row in rows[1:-1]:
                yield l_arrow + pretty_row(row) + r_arrow
                yield border

        else:
            for row in rows[1:4]:
                yield l_arrow + pretty_row(row) + r_arrow
                yield border

            omit_msg = f"..omitting {nrows - 10} rows.."
            omit_gap = ncols // 2 - len(omit_msg) // 2
            omit_pad = "".join(" " for _ in range(omit_gap))

            yield Text(omit_pad + omit_msg, style=dim)
            yield border

            for row in rows[-4:-1]:
                yield l_arrow + pretty_row(row) + r_arrow
                yield border

        yield l_arrow + pretty_row(rows[-1]) + r_border
        yield pad + make_hline(len(rows[-1]) + 2 * gap)



from typing import Any
from typing import Union

import numpy as np

__all__ = ["Face", "Integer", "Float", "Complex", "Bool", "Void"]

Face = Any

_NumpyInt = Union[
    np.int_,
    np.intc,
    np.intp,
    np.int8,
    np.int16,
    np.int32,
    np.int64,
    np.uint8,
    np.uint16,
    np.uint32,
    np.uint64,
]
_NumpyFloat = Union[np.float_, np.float16, np.float32, np.float64]
_NumpyComplex = Union[np.complex_, np.complex64, np.complex128]

Integer = Union[int, _NumpyInt]
Float = Union[float, _NumpyFloat]
Complex = Union[complex, _NumpyComplex]
Bool = Union[bool, np.bool_]
Void = Union[None, np.void]



from pytest import mark

from coinflip import randtests

from .examples import *


@mark.parametrize(example_fields, examples)
def test_examples(randtest, bits, statistic_expect, p_expect, kwargs):
    randtest_method = getattr(randtests, randtest)

    result = randtest_method(bits, **kwargs)

    assert_statistic(result.statistic, statistic_expect)
    assert_p(result.p, p_expect)


@mark.parametrize(multi_example_fields, multi_examples)
def test_multi_examples(randtest, bits, expected_statistics, expected_pvalues, kwargs):
    randtest_method = getattr(randtests, randtest)

    results = randtest_method(bits, **kwargs)

    assert_statistics(results.statistics, expected_statistics)
    assert_pvalues(results.pvalues, expected_pvalues)


@mark.parametrize(sub_example_fields, sub_examples)
def test_sub_examples(randtest, key, bits, statistic_expect, p_expect, kwargs):
    randtest_method = getattr(randtests, randtest)

    meta_result = randtest_method(bits, **kwargs)
    result = meta_result.results[key]

    assert_statistic(result.statistic, statistic_expect)
    assert_p(result.p, p_expect)



from math import ceil
from numbers import Real
from random import getrandbits
from typing import List

from hypothesis import strategies as st
from hypothesis.strategies import SearchStrategy
from typing_extensions import Literal

__all__ = ["mixedbits", "real"]


@st.composite
def mixedbits(draw, min_size=2) -> SearchStrategy[List[Literal[0, 1]]]:
    """Strategy to generate binary sequences"""
    n = draw(st.integers(min_value=min_size, max_value=1000))

    mixedbits = [getrandbits(1) for _ in range(n)]  # TODO make this reproducible
    mixedbits[0:2] = [0, 1]  # force bits being mixed TODO use a filter

    return mixedbits


def real(min_value: Real = None, allow_infinity: bool = True) -> SearchStrategy[Real]:
    return st.one_of(
        st.integers(min_value=ceil(min_value) if min_value else None),
        st.floats(
            min_value=float(min_value) if min_value else None,
            allow_infinity=allow_infinity,
            allow_nan=False,
        ),
    )



"""Utility methods for randomness tests."""
from typing import Any
from typing import Iterator
from typing import Tuple

import pandas as pd

__all__ = [
    "blocks",
    "rawblocks",
    "slider",
]


def blocks(series, blocksize, truncate=True) -> Iterator[pd.Series]:
    """Chunking method for ``Series`` objects

    Parameters
    ----------
    series : ``Series``
        The pandas ``Series`` to chunk
    blocksize : ``Integer``
        Size of the chunks
    truncate : ``bool``, default ``True``
        Whether to discard remaning series

    Yields
    ------
    block : ``Series``
        Chunk of the passed ``series``

    Raises
    ------
    ValueError
        When neither ``blocksize`` or ``nblocks`` is passed
    """
    n = len(series)
    nblocks = n // blocksize
    boundary = blocksize * nblocks

    for i in range(0, boundary, blocksize):
        yield series[i : i + blocksize]

    if not truncate and boundary < n:
        yield series[boundary:]


def rawblocks(*args, **kwargs) -> Iterator[Tuple[Any]]:
    """Tuple chunking method for ``Series`` objects

    Parameters
    ----------
    *args
        Positional arguments to pass to ``blocks``
    **kwargs
        Keyword arguments to pass to ``blocks``

    Yields
    ------
    block_tup : ``Tuple``
        Tuple representation of the block

    Raises
    ------
    ValueError
        When neither ``blocksize`` or ``nblocks`` is passed

    See Also
    --------
    blocks: The method ``rawblocks`` adapts
    """

    for block in blocks(*args, **kwargs):
        block_tup = tuple(block)

        yield block_tup


def slider(series, window_size) -> Iterator[Tuple[Any]]:
    if window_size != 0:
        boundary = len(series) - window_size + 1
        for pointer in range(0, boundary, 1):
            window = series[pointer : pointer + window_size]
            window_tup = tuple(window)

            yield window_tup



from collections import Counter
from dataclasses import dataclass
from math import erfc
from math import sqrt

import pandas as pd
from rich.text import Text
from scipy.stats import chisquare

from coinflip._randtests.common.collections import Bins
from coinflip._randtests.common.core import *
from coinflip._randtests.common.result import MultiTestResult
from coinflip._randtests.common.result import SubTestResult
from coinflip._randtests.common.typing import Integer

__all__ = ["random_excursions", "random_excursions_variant"]

# ------------------------------------------------------------------------------
# Random Excursions Test

# TODO comment why these states and df were chosen
states = [-4, -3, -2, -1, 1, 2, 3, 4]
df = 5

# TODO calculate these dynamically
state_probabilities = {
    1: [0.5000, 0.2500, 0.1250, 0.0625, 0.0312, 0.0312],
    2: [0.7500, 0.0625, 0.0469, 0.0352, 0.0264, 0.0791],
    3: [0.8333, 0.0278, 0.0231, 0.0193, 0.0161, 0.0804],
    4: [0.8750, 0.0156, 0.0137, 0.0120, 0.0105, 0.0733],
    5: [0.9000, 0.0100, 0.0090, 0.0081, 0.0073, 0.0656],
    6: [0.9167, 0.0069, 0.0064, 0.0058, 0.0053, 0.0588],
    7: [0.9286, 0.0051, 0.0047, 0.0044, 0.0041, 0.0531],
}


@randtest()
def random_excursions(series, heads, tails, ctx):
    n = len(series)

    set_task_total(ctx, 4)

    failures = check_recommendations(ctx, {"n ≥ 1000000": n >= 1000000})

    oscillations = series.map({heads: 1, tails: -1})
    cumulative_sums = oscillations.cumsum()

    advance_task(ctx)

    head = pd.Series({-1: 0})
    tail = pd.Series({n + 1: 0})
    walk = pd.concat([head, cumulative_sums, tail])

    advance_task(ctx)

    # TODO standardise or differentiate language of "bins"/"occurences"/"bincounts"
    ncycles = 0
    state_count_bins = {state: Bins(range(df + 1)) for state in states}
    for cycle in ascycles(walk):
        ncycles += 1
        state_counts = Counter(cycle)
        for state in states:
            count = state_counts[state]
            state_count_bins[state][count] += 1

    advance_task(ctx)

    results = {}
    for state in states:
        probabilities = state_probabilities[abs(state)]
        expected_bincounts = [ncycles * prob for prob in probabilities]

        bincounts = state_count_bins[state].values()

        chi2, p = chisquare(list(bincounts), expected_bincounts)

        results[state] = RandomExcursionsSubTestResult(chi2, p, state)

    advance_task(ctx)

    return RandomExcursionsMultiTestResult(heads, tails, failures, results)


@dataclass
class RandomExcursionsSubTestResult(SubTestResult):
    state: Integer


class RandomExcursionsMultiTestResult(MultiTestResult):
    def _pretty_feature(self, result: RandomExcursionsSubTestResult):
        return Text(str(result.state), style="bold")

    def _render(self):
        yield self._results_table("state", "χ²")

    def _render_sub(self, result: RandomExcursionsSubTestResult):
        yield result._pretty_result("chi-square")


def ascycles(walk):
    items = walk.items()
    _, firstval = next(items)
    cycle = [firstval]
    for _, cusum in items:
        if cusum == 0:
            cycle.append(cusum)
            yield cycle
            cycle = [cusum]
        else:
            cycle.append(cusum)


# ------------------------------------------------------------------------------
# Random Excursions Variant Test

variant_states = [-9, -8, -7, -6, -5, -4, -3, -2, -1, 1, 2, 3, 4, 5, 6, 7, 8, 9]


@randtest()
def random_excursions_variant(series, heads, tails, ctx):
    n = len(series)

    set_task_total(ctx, 4)

    failures = check_recommendations(ctx, {"n ≥ 1000000": n >= 1000000})

    oscillations = series.map({heads: 1, tails: -1})
    cumulative_sums = oscillations.cumsum()

    advance_task(ctx)

    head = pd.Series({-1: 0})
    tail = pd.Series({n + 1: 0})
    walk = pd.concat([head, cumulative_sums, tail])

    advance_task(ctx)

    state_counts = walk.value_counts()
    ncycles = state_counts.at[0] - 1

    advance_task(ctx)

    results = {}
    for state in variant_states:
        try:
            count = state_counts.at[state]
        except KeyError:
            count = 0

        p = erfc(abs(count - ncycles) / sqrt(2 * ncycles * (4 * abs(state) - 2)))

        advance_task(ctx)

        results[state] = RandomExcursionsVariantSubTestResult(count, p, state)

    advance_task(ctx)

    return RandomExcursionsVariantMultiTestResult(heads, tails, failures, results)


@dataclass
class RandomExcursionsVariantSubTestResult(SubTestResult):
    state: Integer


class RandomExcursionsVariantMultiTestResult(MultiTestResult):
    def _pretty_feature(self, result: RandomExcursionsVariantSubTestResult):
        return Text(str(result.state), style="bold")

    def _render(self):
        yield self._results_table("state", "ξ")

    def _render_sub(self, result: RandomExcursionsVariantSubTestResult):
        yield result._pretty_result("count")



from collections import Counter
from collections import defaultdict
from itertools import accumulate
from itertools import product
from math import erfc
from math import floor
from math import log
from math import log2
from math import sqrt
from typing import Iterator
from typing import List
from typing import Tuple

import numpy as np
from more_itertools import chunked
from more_itertools import split_at
from more_itertools import windowed
from scipy.fft import fft
from scipy.special import gammaincc
from scipy.stats import chisquare
from scipy.stats import norm
from typing_extensions import Literal

from coinflip.algorithms import berlekamp_massey
from coinflip.algorithms import matrix_rank
from coinflip.collections import Bins
from coinflip.collections import FloorDict
from coinflip.collections import defaultlist

__all__ = [
    "monobit",
    "frequency_within_block",
    "runs",
    "longest_runs",
    "binary_matrix_rank",
    "spectral",
    "non_overlapping_template_matching",
    "overlapping_template_matching",
    "maurers_universal",
    "linear_complexity",
    "serial",
    "approximate_entropy",
    "cusum",
    "random_excursions",
    "random_excursions_variant",
]


def monobit(sequence: List[Literal[0, 1]]):
    n = len(sequence)

    ones = sum(1 for bit in sequence if bit == 1)
    zeroes = n - ones
    diff = abs(ones - zeroes)

    normdiff = diff / sqrt(n)
    p = erfc(normdiff / sqrt(2))

    return normdiff, p


def frequency_within_block(sequence: List[Literal[0, 1]], blocksize: int):
    n = len(sequence)
    nblocks = n // blocksize

    deviations = []
    for block in chunked(sequence, blocksize):
        ones = sum(1 for bit in block if bit == 1)
        prop = ones / blocksize
        dev = prop - 1 / 2
        deviations.append(dev)

    chi2 = 4 * blocksize * sum(x ** 2 for x in deviations)
    p = gammaincc(nblocks / 2, chi2 / 2)

    return chi2, p


def runs(sequence: List[Literal[0, 1]]):
    n = len(sequence)

    ones = sum(1 for bit in sequence if bit == 1)

    prop_ones = ones / n
    prop_zeroes = 1 - prop_ones

    nruns = sum(1 for _ in asruns(sequence))
    p = erfc(
        abs(nruns - (2 * ones * prop_zeroes))
        / (2 * sqrt(2 * n) * prop_ones * prop_zeroes)
    )

    return nruns, p


def longest_runs(sequence: List[Literal[0, 1]]):
    n_defaults = FloorDict(
        {
            128: (8, 16, [1, 2, 3, 4]),
            6272: (128, 49, [4, 5, 6, 7, 8, 9]),
            750000: (10 ** 4, 75, [10, 11, 12, 13, 14, 15, 16]),
        }
    )
    n = len(sequence)
    blocksize, nblocks, intervals = n_defaults[n]
    maxlen_bins = Bins(intervals)

    blocksize_probabilities = {
        8: [0.2148, 0.3672, 0.2305, 0.1875],
        128: [0.1174, 0.2430, 0.2493, 0.1752, 0.1027, 0.1124],
        512: [0.1170, 0.2460, 0.2523, 0.1755, 0.1027, 0.1124],
        1000: [0.1307, 0.2437, 0.2452, 0.1714, 0.1002, 0.1088],
        10000: [0.0882, 0.2092, 0.2483, 0.1933, 0.1208, 0.0675, 0.0727],
    }
    probabilities = blocksize_probabilities[blocksize]
    expected_bincounts = [prob * nblocks for prob in probabilities]

    boundary = nblocks * blocksize
    for block in chunked(sequence[:boundary], blocksize):
        one_run_lengths = [length for value, length in asruns(block) if value == 1]

        try:
            maxlen = max(one_run_lengths)
        except ValueError:
            maxlen = 0

        maxlen_bins[maxlen] += 1

    chi2, p = chisquare(list(maxlen_bins.values()), expected_bincounts)

    return chi2, p


def binary_matrix_rank(sequence: List[Literal[0, 1]], matrix_dimen: Tuple[int, int]):
    n = len(sequence)
    nrows, ncols = matrix_dimen
    blocksize = nrows * ncols
    nblocks = n // blocksize
    trunc_sequence = sequence[: nblocks * blocksize]

    expected_rankcounts = (0.2888 * nblocks, 0.5776 * nblocks, 0.1336 * nblocks)

    ranks = []
    for block in chunked(trunc_sequence, blocksize):
        matrix = chunked(block, ncols)
        rank = matrix_rank(matrix)
        ranks.append(rank)

    fullrank = min(nrows, ncols)
    rankcounts = [0 for _ in range(3)]
    for rank in ranks:
        if rank == fullrank:
            rankcounts[0] += 1
        elif rank == fullrank - 1:
            rankcounts[1] += 1
        else:
            rankcounts[2] += 1

    chi2, p = chisquare(rankcounts, expected_rankcounts)

    return chi2, p


def spectral(sequence: List[Literal[0, 1]]):
    n = len(sequence)
    if n % 2 != 0:
        sequence = sequence[:-1]

    threshold = sqrt(log(1 / 0.05) * n)
    nbelow_expect = 0.95 * n / 2

    oscillations = [bit if bit == 1 else -1 for bit in sequence]
    fourier = fft(oscillations)
    half_fourier = fourier[: n // 2]

    peaks = [abs(complex_num) for complex_num in half_fourier]
    nbelow = sum(peak < threshold for peak in peaks)

    diff = nbelow - nbelow_expect
    normdiff = diff / sqrt((n * 0.95 * 0.05) / 4)
    p = erfc(abs(normdiff) / sqrt(2))

    return normdiff, p


def non_overlapping_template_matching(
    sequence: List[Literal[0, 1]], template_size: int, blocksize: int
):
    n = len(sequence)
    nblocks = n // blocksize

    matches_expect = (blocksize - template_size + 1) / 2 ** template_size
    variance = blocksize * (
        (1 / 2 ** template_size) - ((2 * template_size - 1)) / 2 ** (2 * template_size)
    )

    template_block_matches = defaultdict(lambda: defaultlist(int))
    for i, block in enumerate(chunked(sequence, blocksize)):
        matches = defaultdict(int)

        for window in windowed(block, template_size):
            matches[window] += 1

        for template, matches in matches.items():
            template_block_matches[template][i] = matches

    statistics = []
    pvalues = []
    for template in product([0, 1], repeat=template_size):
        block_matches = template_block_matches[template][:nblocks]
        match_diffs = [matches - matches_expect for matches in block_matches]

        chi2 = sum(diff ** 2 / variance for diff in match_diffs)
        p = gammaincc(nblocks / 2, chi2 / 2)

        statistics.append(chi2)
        pvalues.append(p)

    return statistics, pvalues


def overlapping_template_matching(
    sequence: List[Literal[0, 1]], template_size: int, blocksize: int
):
    n = len(sequence)
    template = [1 for _ in range(template_size)]
    nblocks = n // blocksize
    trunc_sequence = sequence[: nblocks * blocksize]

    probabilities = [0.367879, 0.183939, 0.137954, 0.099634, 0.069935, 0.140656]
    expected_tallies = [prob * nblocks for prob in probabilities]

    block_matches = []
    for block in chunked(trunc_sequence, blocksize):
        matches = 0

        for window in windowed(block, template_size):
            if all(wbit == tbit for wbit, tbit in zip(window, template)):
                matches += 1

        block_matches.append(matches)

    tallies = [0 for _ in range(6)]
    for matches in block_matches:
        i = min(matches, 5)
        tallies[i] += 1

    chi2, p = chisquare(tallies, expected_tallies)

    return chi2, p


def maurers_universal(sequence: List[Literal[0, 1]], blocksize: int, init_nblocks: int):
    n = len(sequence)
    nblocks = n // blocksize
    segment_nblocks = nblocks - init_nblocks

    boundary = init_nblocks * blocksize
    init_sequence = sequence[:boundary]
    test_sequence = sequence[boundary : nblocks * blocksize]

    blocksize_dists = {
        1: (0.7326495, 0.690),
        2: (1.5374383, 1.338),
        3: (2.4016068, 1.901),
        4: (3.3112247, 2.358),
        5: (4.2534266, 2.705),
        6: (5.2177052, 2.954),
        7: (6.1962507, 3.125),
        8: (7.1836656, 3.238),
        9: (8.1764248, 3.311),
        10: (9.1723243, 3.356),
        11: (10.170032, 3.384),
        12: (11.168765, 3.401),
        13: (12.168070, 3.410),
        14: (13.167693, 3.416),
        15: (14.167488, 3.419),
        16: (15.167379, 3.421),
    }
    mean_expect, variance = blocksize_dists[blocksize]

    last_permutation_pos = defaultdict(int)
    for pos, permutation in enumerate(chunked(init_sequence, blocksize), 1):
        last_permutation_pos[tuple(permutation)] = pos

    log2_distances = 0
    for pos, permutation in enumerate(
        chunked(test_sequence, blocksize), init_nblocks + 1
    ):
        distance = pos - last_permutation_pos[tuple(permutation)]
        log2_distances += log2(distance)

        last_permutation_pos[tuple(permutation)] = pos

    norm_distances = log2_distances / segment_nblocks
    p = erfc(abs((norm_distances - mean_expect) / (sqrt(2 * variance))))

    return norm_distances, p


def serial(sequence: List[Literal[0, 1]], blocksize: int):
    n = len(sequence)

    normalised_sums = {}
    for window_size in [blocksize, blocksize - 1, blocksize - 2]:
        if window_size > 0:
            head = sequence[: window_size - 1]
            ouroboros = sequence + head

            counts = defaultdict(int)
            for window in windowed(ouroboros, window_size):
                counts[tuple(window)] += 1

            sum_squares = sum(count ** 2 for count in counts.values())
            normsum = (2 ** window_size / n) * sum_squares - n

        else:
            normsum = 0

        normalised_sums[window_size] = normsum

    normsum_deltas = [
        normalised_sums[blocksize] - normalised_sums[blocksize - 1],
        normalised_sums[blocksize]
        - 2 * normalised_sums[blocksize - 1]
        + normalised_sums[blocksize - 2],
    ]

    pvalues = [
        gammaincc(2 ** (blocksize - 2), normsum_deltas[0] / 2),
        gammaincc(2 ** (blocksize - 3), normsum_deltas[1] / 2),
    ]

    return normsum_deltas, pvalues


def linear_complexity(sequence: List[Literal[0, 1]], blocksize: int):
    n = len(sequence)
    nblocks = n // blocksize
    trunc_sequence = sequence[: nblocks * blocksize]

    probabilities = [0.010417, 0.03125, 0.125, 0.5, 0.25, 0.0625, 0.020833]
    mean_expect = (
        blocksize / 2
        + (9 + (-(1 ** (blocksize + 1)))) / 36
        - (blocksize / 3 + 2 / 9) / 2 ** blocksize
    )
    expected_bincounts = [nblocks * prob for prob in probabilities]

    variance_bins = Bins([-3, -2, -1, 0, 1, 2, 3])
    for block in chunked(trunc_sequence, blocksize):
        linear_complexity = berlekamp_massey(block)
        variance = (-1) ** blocksize * (linear_complexity - mean_expect) + 2 / 9
        variance_bins[variance] += 1

    chi2, p = chisquare(list(variance_bins.values()), expected_bincounts)

    return chi2, p


def approximate_entropy(sequence: List[Literal[0, 1]], blocksize: int):
    n = len(sequence)

    phis = []
    for template_size in [blocksize, blocksize + 1]:
        head = sequence[: template_size - 1]
        ouroboros = sequence + head

        permutation_counts = defaultdict(int)
        for window in windowed(ouroboros, template_size):
            permutation_counts[tuple(window)] += 1

        normalised_counts = []
        for count in permutation_counts.values():
            normcount = count / n
            normalised_counts.append(normcount)

        phi = sum(normcount * log(normcount) for normcount in normalised_counts)
        phis.append(phi)

    approx_entropy = phis[0] - phis[1]
    chi2 = 2 * n * (log(2) - approx_entropy)
    p = gammaincc(2 ** (blocksize - 1), chi2 / 2)

    return chi2, p


def cusum(sequence: List[Literal[0, 1]], reverse: bool = False):
    n = len(sequence)

    oscillations = [bit if bit == 1 else -1 for bit in sequence]
    if reverse:
        oscillations = oscillations[::-1]
    cusums = accumulate(oscillations)

    abs_cusums = [abs(cusum) for cusum in cusums]
    max_cusum = max(abs_cusums)

    p = (
        1
        - sum(
            norm.cdf((4 * k + 1) * max_cusum / sqrt(n))
            - norm.cdf((4 * k - 1) * max_cusum / sqrt(n))
            for k in np.arange(
                floor((-n / max_cusum + 1) / 4), floor((n / max_cusum - 1) / 4) + 1, 1
            )
        )
        + sum(
            norm.cdf((4 * k + 3) * max_cusum / sqrt(n))
            - norm.cdf((4 * k + 1) * max_cusum / sqrt(n))
            for k in np.arange(
                floor((-n / max_cusum - 3) / 4), floor((n / max_cusum - 1) / 4) + 1, 1
            )
        )
    )

    return max_cusum, p


def random_excursions(sequence: List[Literal[0, 1]]):
    states = [-4, -3, -2, -1, 1, 2, 3, 4]
    state_count_bins = {state: Bins(range(6)) for state in states}

    state_probabilities = {
        1: [0.5000, 0.2500, 0.1250, 0.0625, 0.0312, 0.0312],
        2: [0.7500, 0.0625, 0.0469, 0.0352, 0.0264, 0.0791],
        3: [0.8333, 0.0278, 0.0231, 0.0193, 0.0161, 0.0804],
        4: [0.8750, 0.0156, 0.0137, 0.0120, 0.0105, 0.0733],
        5: [0.9000, 0.0100, 0.0090, 0.0081, 0.0073, 0.0656],
        6: [0.9167, 0.0069, 0.0064, 0.0058, 0.0053, 0.0588],
        7: [0.9286, 0.0051, 0.0047, 0.0044, 0.0041, 0.0531],
    }

    oscillations = [bit if bit == 1 else -1 for bit in sequence]
    cusums = accumulate(oscillations)

    ncycles = 0
    for cycle in split_at(cusums, lambda x: x == 0):
        ncycles += 1

        counts = Counter(cycle)
        for state in states:
            count = counts[state]
            state_count_bins[state][count] += 1

    statistics = []
    pvalues = []
    for state in states:
        probabilities = state_probabilities[abs(state)]
        expected_bincounts = [ncycles * prob for prob in probabilities]

        bincounts = state_count_bins[state].values()

        chi2, p = chisquare(list(bincounts), expected_bincounts)

        statistics.append(chi2)
        pvalues.append(p)

    return statistics, pvalues


def random_excursions_variant(sequence: List[Literal[0, 1]]):
    states = [-9, -8, -7, -6, -5, -4, -3, -2, -1, 1, 2, 3, 4, 5, 6, 7, 8, 9]

    oscillations = [bit if bit == 1 else -1 for bit in sequence]
    cusums = accumulate(oscillations)

    state_counts = Counter(cusums)
    ncycles = state_counts[0] + 1

    counts = []
    pvalues = []
    for state in states:
        count = state_counts[state]
        p = erfc(abs(count - ncycles) / sqrt(2 * ncycles * (4 * abs(state) - 2)))

        counts.append(count)
        pvalues.append(p)

    return counts, pvalues


# ------------------------------------------------------------------------------
# Helpers


def asruns(sequence: List[Literal[0, 1]]) -> Iterator[Tuple[Literal[0, 1], int]]:
    run_val = sequence[0]
    run_len = 1

    for value in sequence[1:]:
        if value == run_val:
            run_len += 1

        else:
            yield run_val, run_len

            run_val = value
            run_len = 1

    yield run_val, run_len



import pandas as pd
from pytest import raises

from coinflip._randtests.common.testutils import slider


def test_slider():
    series = pd.Series([0, 1, 0, 0, 1])
    it = slider(series, 0)

    with raises(StopIteration):
        next(it)



from dataclasses import dataclass

import pandas as pd

from coinflip._randtests.common.exceptions import NonBinarySequenceError

__all__ = ["DataParsingError", "parse_text", "parse_binary"]


class DataParsingError(ValueError):
    """Base class for parsing-related errors"""


@dataclass
class MultipleColumnsError(DataParsingError):
    """Error for when only one column of data was expected"""

    ncols: int

    def __str__(self):
        return (
            f"Parsed data contains {self.ncols} columns, but only 1 column was expected"
        )


def parse_text(data_file) -> pd.Series:
    """Reads file containing data into a pandas Series

    Reads from file containing RNG output and produces a representitive pandas
    Series. The appropiate dtype is inferred from the data itself.

    Parameters
    ----------
    data_file : file-like object
        File containing RNG output

    Returns
    -------
    ``Series``
        A pandas ``Series`` which represents the data

    Raises
    ------
    MultipleColumnsError
        If inputted data contains multiple values per line
    NonBinarySequenceError
        If sequence does not contain only 2 values

    See Also
    --------
    pandas.read_csv : The pandas method for reading ``data_file``
    """
    df = pd.read_csv(data_file, header=None)

    ncols = len(df.columns)
    if ncols > 1:
        raise MultipleColumnsError(ncols)
    series = df.iloc[:, 0]

    if series.nunique() != 2:
        raise NonBinarySequenceError()

    series = series.infer_objects()

    return series


def parse_binary(data_file) -> pd.Series:
    sequence = []
    with open(data_file, "rb") as f:
        bytes_ = f.read()
        for byte in bytes_:
            bitstring = format(byte, "08b")
            bits = [int(bit) for bit in bitstring]

            sequence += bits

    series = pd.Series(sequence)

    return series



from dataclasses import astuple
from dataclasses import dataclass
from math import floor
from math import sqrt
from typing import Iterable
from typing import Tuple

from scipy.stats import chisquare
from typing_extensions import Literal

from coinflip._randtests.common.core import *
from coinflip._randtests.common.result import TestResult
from coinflip._randtests.common.result import make_chisquare_table
from coinflip._randtests.common.testutils import blocks
from coinflip._randtests.common.testutils import rawblocks
from coinflip._randtests.common.typing import Integer

__all__ = ["binary_matrix_rank", "matrix_rank"]


@dataclass
class RankCounts:
    full: Integer = 0
    runnerup: Integer = 0
    remaining: Integer = 0


@randtest(min_n=4)
def binary_matrix_rank(
    series, heads, tails, ctx, matrix_dimen: Tuple[Integer, Integer] = None
):
    n = len(series)

    if matrix_dimen is None:
        if n // (32 * 32) > 38:
            nrows = 32
            ncols = 32
        else:
            blocksize = max(n // 38, 4)
            nrows = floor(sqrt(blocksize))
            ncols = nrows
    else:
        nrows, ncols = matrix_dimen

    blocksize = nrows * ncols
    nblocks = n // blocksize

    set_task_total(ctx, nblocks + 3)

    failures = check_recommendations(
        ctx,
        {
            "n ≥ 128": n >= 152,  # nblocks=38, blocksize=4
            "n ≥ 38 * blocksize": n >= 38 * blocksize,
        },
    )

    fullrank = min(nrows, ncols)

    # TODO find expressive and performative calculation for constants
    expected_rankcounts = RankCounts(
        full=0.2888 * nblocks, runnerup=0.5776 * nblocks, remaining=0.1336 * nblocks,
    )

    rankable_series = series.map({heads: 1, tails: 0})

    advance_task(ctx)

    ranks = []
    for block in blocks(rankable_series, blocksize):
        matrix = [row for row in rawblocks(block, ncols)]

        rank = matrix_rank(matrix)

        advance_task(ctx)

        ranks.append(rank)

    rankcounts = RankCounts()
    for rank in ranks:
        if rank == fullrank:
            rankcounts.full += 1
        elif rank == fullrank - 1:
            rankcounts.runnerup += 1
        else:
            rankcounts.remaining += 1

    advance_task(ctx)

    statistic, p = chisquare(astuple(rankcounts), astuple(expected_rankcounts))

    advance_task(ctx)

    return BinaryMatrixRankTestResult(
        heads,
        tails,
        failures,
        statistic,
        p,
        nrows,
        ncols,
        fullrank,
        expected_rankcounts,
        rankcounts,
    )


@dataclass
class BinaryMatrixRankTestResult(TestResult):
    nrows: Integer
    ncols: Integer
    fullrank: Integer
    expected_rankcounts: RankCounts
    rankcounts: RankCounts

    def _render(self):
        yield self._pretty_result("chi-square")

        yield TestResult._pretty_inputs(
            ("nrows", self.nrows), ("ncols", self.ncols),
        )

        runnerup = self.fullrank - 1
        remaining = runnerup - 1
        f_ranks = [
            str(self.fullrank),
            str(runnerup),
            "0" if remaining == 0 else f"0-{remaining}",
        ]

        table = make_chisquare_table(
            "rank of matrix",
            "ranks",
            f_ranks,
            astuple(self.expected_rankcounts),
            astuple(self.rankcounts),
        )

        yield table


def matrix_rank(matrix: Iterable[Iterable[Literal[0, 1]]]) -> Integer:
    """Finds the rank of a binary matrix

    Parameters
    ----------
    matrix : ``Iterable[Iterable[Integer]]``
        Binary matrix to rank

    Returns
    -------
    rank : Integer
        Rank of ``matrix``

    Notes
    -----
    Implementaton inpisred by a `StackOverflow answer
    <https://stackoverflow.com/a/56858995/5193926>`_ from Mark Dickinson.
    """
    numbers = [bits2int(bits) for bits in matrix]

    rank = 0
    while len(numbers) > 0:
        pivot = numbers.pop()
        if pivot:
            rank += 1
            lsb = pivot & -pivot
            for i, num in enumerate(numbers):
                if lsb & num:
                    numbers[i] = num ^ pivot

    return rank


def bits2int(bits: Iterable[Literal[0, 1]]) -> Integer:
    """Converts a list of bits into a numerical representation"""
    num = 0
    for bit in bits:
        num = (num << 1) | bit

    return num



from collections import defaultdict
from dataclasses import dataclass
from math import floor
from math import log2
from typing import DefaultDict
from typing import Dict
from typing import Tuple

import pandas as pd
from rich.table import Table
from rich.text import Text
from scipy.special import gammaincc

from coinflip._randtests.common.core import *
from coinflip._randtests.common.result import MultiTestResult
from coinflip._randtests.common.result import SubTestResult
from coinflip._randtests.common.testutils import slider
from coinflip._randtests.common.typing import Face
from coinflip._randtests.common.typing import Float
from coinflip._randtests.common.typing import Integer

__all__ = ["serial"]


@randtest()
def serial(series, heads, tails, ctx, blocksize=None):
    n = len(series)

    if not blocksize:
        blocksize = max(floor(log2(n)) - 2 - 1, 2)

    set_task_total(ctx, (1 + n) * 3 + 2)

    failures = check_recommendations(
        ctx, {"blocksize < ⌊log2(n) - 2⌋": blocksize < floor(log2(n)) - 2}
    )

    permutation_counts = {}
    normalised_sums = {}
    for window_size in [blocksize, blocksize - 1, blocksize - 2]:
        if window_size > 0:
            head = series[: window_size - 1]
            ouroboros = pd.concat([series, head])

            advance_task(ctx)

            counts = defaultdict(int)
            for window_tup in slider(ouroboros, window_size):
                counts[window_tup] += 1

                advance_task(ctx)

            permutation_counts[window_size] = counts

            sum_squares = sum(count ** 2 for count in counts.values())
            normsum = (2 ** window_size / n) * sum_squares - n

            normalised_sums[window_size] = normsum

        else:
            permutation_counts[window_size] = defaultdict(int)
            normalised_sums[window_size] = 0

    advance_task(ctx)

    normsum_delta1 = normalised_sums[blocksize] - normalised_sums[blocksize - 1]
    p1 = gammaincc(2 ** (blocksize - 2), normsum_delta1 / 2)

    normsum_delta2 = (
        normalised_sums[blocksize]
        - 2 * normalised_sums[blocksize - 1]
        + normalised_sums[blocksize - 2]
    )
    p2 = gammaincc(2 ** (blocksize - 3), normsum_delta2 / 2)

    advance_task(ctx)

    results = {
        "∇ψ²ₘ": SubTestResult(normsum_delta1, p1,),
        "∇²ψ²ₘ": SubTestResult(normsum_delta2, p2,),
    }

    return SerialMultiTestResult(
        heads, tails, failures, results, blocksize, permutation_counts, normalised_sums,
    )


@dataclass
class SerialMultiTestResult(MultiTestResult):
    blocksize: Integer
    permutation_counts: Dict[Integer, DefaultDict[Tuple[Face, ...], Integer]]
    normalised_sums: Dict[Integer, Float]

    def _render(self):
        first, second = list(self.results.items())

        f_first = first[1]._pretty_result(prefix="first", stat_varname="delta psi²")
        f_second = second[1]._pretty_result(prefix="second", stat_varname="delta² psi²")

        yield f_first
        yield f_second

        yield self._pretty_inputs(("blocksize", self.blocksize),)

    @classmethod
    def _render_single(
        cls, feature: str, feature_notation: str, result: SubTestResult
    ) -> Table:
        grid = Table.grid(padding=(1))

        title = Text(feature, style="bold")
        grid.add_row(title)

        grid.add_row(f_result)

        return grid



from pytest import mark

from coinflip import generators
from coinflip.cli.runner import list_tests

# TODO make this deterministic
RNG = generators.python()
bits = [next(RNG) for _ in range(1000000)]

randtests = [(func,) for _, func in list_tests()]


@mark.parametrize(["randtest"], randtests)
@mark.timeout(1)
def test_thousand_bits(randtest):
    randtest(bits[:1000])


@mark.slow
@mark.parametrize(["randtest"], randtests)
@mark.timeout(120)  # 2 minutes
def test_million_bits(randtest):
    randtest(bits)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['coinflip',
 'coinflip._randtests',
 'coinflip._randtests.common',
 'coinflip.cli']

package_data = \
{'': ['*']}

install_requires = \
['Jinja2>=2.11.2,<3.0.0',
 'altair>=4.1.0,<5.0.0',
 'click>=7.1.2,<8.0.0',
 'numpy>=1.19.4,<2.0.0',
 'pandas>=1.1.4,<2.0.0',
 'rich>=9.2.0,<10.0.0',
 'scipy>=1.5.4,<2.0.0',
 'typing-extensions>=3.7.4,<4.0.0']

extras_require = \
{'docs': ['sphinx>=3,<4',
          'sphinx-rtd-theme>=0.5.0,<0.6.0',
          'sphinx-click>=2.5.0,<3.0.0'],
 'refimpl': ['more-itertools>=8.6.0,<9.0.0'],
 'test': ['tox>=3.20.1,<4.0.0',
          'pytest>=6.1.2,<7.0.0',
          'hypothesis>=5.41.2,<6.0.0',
          'pytest-timeout>=1.4.2,<2.0.0']}

entry_points = \
{'console_scripts': ['coinflip = coinflip.cli.commands:main']}

setup_kwargs = {
    'name': 'coinflip',
    'version': '0.1.2',
    'description': 'Randomness testing for humans',
    'long_description': '========\ncoinflip\n========\n\nRandomness testing for humans\n\n|build| |docs| |license| |version| |supported-versions| |hypothesis| |black|\n\n*coinflip* aims to implement the tests recommended by `NIST SP800-22\n<https://csrc.nist.gov/publications/detail/sp/800-22/rev-1a/final>`_\nto check random number generators for randomness.  A user-friendly command-line\ninterface provided allows you to ``run`` the tests on your data, and\nsubsequently ``report`` on the results by generating informational HTML\ndocuments.\n\n``coinflip.randtests`` acts as the `public API\n<https://coinflip.readthedocs.io/en/latest/reference/randtests.html>`_\nfor `notebook users <https://jupyter.org/index.html>`_ and developers to use\nthe randomness tests directly. The tests are implemented as general solutions,\nmeaning they accept basically any sequence with two distinct elements!\n\n.. image:: https://raw.githubusercontent.com/Honno/coinflip/report/.video_files/thumb/thumb_thumb.png\n   :target: http://www.youtube.com/watch?v=0xrWG3Ki9Z8\n\nSetup\n=====\n\nYou can get the latest release of coinflip from PyPI.\n\n.. code-block:: console\n\n    $ pip install coinflip\n\nAlternatively you can get the (unstable) development version straight from\nGitHub.\n\n.. code-block:: console\n\n    $ pip install git+https://github.com/Honno/coinflip\n\nIf that means nothing to you, no fret! Please continue reading the instructions\nbelow.\n\nInstall Python 3.7+\n-------------------\n\nCross-platform installation instructions for Python  are available at\n`realpython.com/installing-python/ <https://realpython.com/installing-python/>`_.\n\nNote ``coinflip`` only works on **Python 3.7 or above**. Make sure you have\nPython 3.7 (or higher) by checking the version of your installation:\n\n.. code-block:: console\n\n    $ python --version\n    Python 3.7.X\n\nClone repository\n----------------\n\nYou can clone the source code via `Git\n<https://www.freecodecamp.org/news/what-is-git-and-how-to-use-it-c341b049ae61/>`_:\n\n.. code-block:: console\n\n    $ git clone https://github.com/Honno/coinflip\n\n\nInstall coinflip\n----------------\n\nEnter the directory *coinflip* is downloaded to:\n\n.. code-block:: console\n\n   $ cd coinflip\n\nYou can install *coinflip* via the ``pip`` module:\n\n.. code-block:: console\n\n    $ pip install -e .\n\n`pip <https://realpython.com/what-is-pip/>`_ is the standard package manager for\nPython, which should of installed automatically when installing Python 3.7+.\n\nTrial run\n---------\n\nTry running the randomness tests on an automatically generated binary sequence:\n\n.. code-block:: console\n\n    $ coinflip example-run\n\nIf the command ``coinflip`` is "not found", you may need to add your local\nbinaries folder to your shell\'s path. For example, in bash you would do the\nfollowing:\n\n.. code-block:: console\n\n    $ echo "export PATH=~/.local/bin:$PATH" >> ~/.bash_profile\n    $ source ~/.bash_profile\n\nIn the worst case, you can execute commands via ``python -m``:\n\n.. code-block:: console\n\n    $ python -m coinflip example-run\n\n\nQuick start\n===========\n\nRandomness tests can be ran over your RNG output via the ``run`` command.\n\n.. code-block:: console\n\n    $ coinflip run DATA OUT\n    ...\n\n``DATA`` is the path to newline-delimited text file that contains a binary\nsequence. An example file to use is available on `my gist\n<https://gist.github.com/Honno/dd6f3527e588428fa17a999042e3c6e8>`_.\nAlternatively, raw binary files can be read as bitstreams via the ``--binary``\nflag\n\n``OUT`` is the path where you want the results to be saved. The results will be\nsaved as a `pickle <https://docs.python.org/3/library/pickle.html>`_-serialised\nfile, which can be viewed again via the ``read`` command. Additionally you can\ngenerate informational HTML reports from the results via the ``report`` command,\nbut note that the reports are currently very lacking.\n\nOutput should comprise of the sequence parsed from ``DATA``, test-specific result\nsummaries, and a final overall summary table.\n\n.. |build| image:: https://img.shields.io/github/workflow/status/Honno/coinflip/Test\n    :target: https://github.com/Honno/coinflip/actions?query=workflow%3ATest\n    :alt: GitHub Workflow Status\n\n.. |docs| image:: https://readthedocs.org/projects/coinflip/badge/?style=flat\n    :target: https://readthedocs.org/projects/coinflip\n    :alt: Documentation Status\n\n.. |hypothesis| image:: https://img.shields.io/badge/hypothesis-tested-brightgreen.svg\n   :alt: Tested with Hypothesis\n   :target: https://hypothesis.readthedocs.io\n\n.. |version| image:: https://img.shields.io/pypi/v/coinflip.svg\n    :alt: PyPI package latest release\n    :target: https://pypi.org/project/coinflip\n\n.. |supported-versions| image:: https://img.shields.io/badge/python-3.7%2B-informational\n    :alt: Supported versions\n    :target: https://pypi.org/project/coinflip\n\n.. |supported-implementations| image:: https://img.shields.io/pypi/implementation/coinflip.svg\n    :alt: Supported implementations\n    :target: https://pypi.org/project/coinflip\n\n.. |black| image:: https://img.shields.io/badge/code%20style-black-000000.svg\n    :alt: Code style\n    :target: https://github.com/psf/black\n\n.. |license| image:: https://img.shields.io/badge/license-BSD-blueviolet\n    :alt: License\n    :target: https://github.com/Honno/coinflip/blob/master/LICENSE.md\n',
    'author': 'Matthew Barber',
    'author_email': 'quitesimplymatt@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/Honno/coinflip',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



from typing import List
from typing import NamedTuple

from pytest import mark
from typing_extensions import Literal

from coinflip.algorithms import berlekamp_massey

__all__ = ["bm_examples"]


class BMExample(NamedTuple):
    sequence: List[Literal[0, 1]]
    min_size: int


bm_examples = [
    BMExample(
        # SP800-22 example
        # 1-4  | 1 1 0 1  | 0
        # 2-5  | 1 0 1 0  | 1
        # 3-6  | 0 1 0 1  | 1
        # 4-7  | 1 0 1 1  | 1
        # 5-8  | 0 1 1 1  | 1
        # 6-9  | 1 1 1 1  | 0
        # 7-10 | 1 1 1 0  | 0
        # 8-11 | 1 1 0 0  | 0
        # 9-12 | 1 0 0 0  | 1
        sequence=[1, 1, 0, 1, 0, 1, 1, 1, 1, 0, 0, 0, 1],
        min_size=4,
    ),
    BMExample(
        # https://bell0bytes.eu/linear-feedback-shift-registers/
        sequence=[
            1,
            0,
            0,
            1,
            1,
            1,
            0,
            1,
            1,
            0,
            0,
            1,
            1,
            1,
            0,
            1,
            0,
            1,
            0,
            0,
            1,
            0,
            0,
            1,
            1,
        ],
        min_size=13,
    ),
]


@mark.parametrize(BMExample._fields, bm_examples)
def test_berlekamp_massey(sequence, min_size):
    assert berlekamp_massey(sequence) == min_size



from dataclasses import dataclass
from functools import lru_cache
from functools import wraps
from typing import Dict
from typing import List
from typing import NamedTuple
from typing import Optional
from typing import Tuple
from warnings import warn

import pandas as pd
from rich.progress import Progress

from coinflip._randtests.common.exceptions import NonBinarySequenceError
from coinflip._randtests.common.exceptions import TestInputError
from coinflip._randtests.common.typing import Face

__all__ = [
    "randtest",
    "set_task_total",
    "advance_task",
    "check_recommendations",
]


@dataclass
class MinimumInputError(TestInputError):
    n: int
    min_n: int

    def __str__(self):
        return f"Sequence length {self.n} is below required minimum of {self.min_n}"


class CliContext(NamedTuple):
    progress: Progress
    task: int


def randtest(min_n=2):
    """Decorator factory for parsing sequences in randomness tests

    Returns a decorator (a method which returns a wrapper method). The wrapper
    checks if passed ``sequence`` is a pandas ``Series``, attempting to convert
    it if not.

    The length of the ``sequence`` is then checked to see if it meets the
    passed minimum input requirement, raising an error if not.

    The `heads` and `tails` of the ``sequence`` is inferred from its unique
    values.

    Parameters
    ----------
    min_n : ``int``, default ``2``
        Minimum length of sequence the test can handle

    Returns
    -------
    decorator : ``Callable[[Callable], Callable]``
        Decorator method for parsing sequences in randomness tests

    Raises
    ------
    MinimumInputError
        If ``sequence`` length exceeds ``min_n``

    See Also
    --------
    pandas.Series: Initialises with ``sequence`` if not already a ``Series``

    Notes
    -----
    `heads` and `tails` corresponds to the ``1`` and ``0`` values of a binary
    input sequence. SP800-22 only specifies randomness tests for these binary
    sequences, but in actuality all of its tests can be generalised for any
    sequence of a GF(2) field (i.e. contains two distinct values), so the
    `heads` and `tails` abstraction allows for mapping these GF(2) sequences to
    their binary counterparts.
    """

    def decorator(func):
        @wraps(func)
        def wrapper(sequence, ctx: Progress = None, **kwargs):
            if isinstance(sequence, pd.Series):
                series = sequence
            else:
                series = pd.Series(sequence)

            if series.nunique() != 2:
                raise NonBinarySequenceError()

            n = len(series)
            if n < min_n:
                raise MinimumInputError(n, min_n)

            values = series.unique()
            heads, tails = infer_faces(tuple(values))

            result = func(series, heads, tails, ctx, **kwargs)

            return result

        return wrapper

    return decorator


@lru_cache()
def infer_faces(unique_values: Tuple[Face, Face]) -> Tuple[Face, Face]:
    """Infers the `heads` and `tails` faces from a list of unique values

    An equality check between the values is attempted where the "largest"
    value is chosen as the `heads`, and subsequently the other value is
    chosen as the `tails`.

    Parameters
    ----------
    unique_values : ``Tuple[Any, Any]``
        Tuple of two unique values

    Returns
    -------
    heads
        Inferred heads face of ``unique_values``
    tails
        Inferred tails face of ``unique_values``
    """
    heads = max(unique_values)
    tails = next(value for value in unique_values if value != heads)

    return heads, tails


def progress_context(func):
    @wraps(func)
    def wrapper(ctx: Optional[CliContext], *args):
        if ctx:
            progress, task = ctx
            func(progress, task, *args)

    return wrapper


@progress_context
def set_task_total(progress, task, total: int):
    progress.update(task, total=total)


@progress_context
def advance_task(progress, task):
    progress.update(task, advance=1)


def check_recommendations(ctx: Optional[CliContext], recommendations: Dict[str, bool]):
    """Warns on recommendation failures

    Parameters
    ----------
    recommendations : ``Dict[str, bool]``
        Map of recommendation string representations to the actual
        recommendation outcomes

    Warns
    -----
    UserWarning
        When one or more recommendations fail
    """
    failures = [expr for expr, success in recommendations.items() if not success]

    if failures and not ctx:
        warn(make_failures_msg(failures), UserWarning)

    return failures


def make_failures_msg(failures: List[str]) -> str:
    nfail = len(failures)

    if nfail == 1:
        expr = failures[0]
        return f"NIST recommendation not met: {expr}"

    elif nfail > 1:
        msg = "Multiple NIST recommendations not met:\n"
        msg += "\n".join([f"  • {expr}" for expr in failures])

        return msg
